# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import os
from proboscis import SkipTest
import re
import tempfile
import time

from troveclient.compat import exceptions

from trove.common import exception
from trove.common.utils import poll_until
from trove.guestagent.common import guestagent_utils
from trove.guestagent.common import operating_system
from trove.module import models
from trove.tests.scenario.runners.test_runners import TestRunner


class ModuleRunner(TestRunner):

    def __init__(self):
        super(ModuleRunner, self).__init__()

        self.MODULE_CONTENTS_PATTERN = 'Message=%s\n'
        self.MODULE_MESSAGE_PATTERN = 'Hello World from: %s'
        self.MODULE_NAME = 'test_module_1'
        self.MODULE_DESC = 'test description'
        self.MODULE_NEG_CONTENTS = 'contents for negative tests'
        self.MODULE_BINARY_SUFFIX = '_bin_auto'
        self.MODULE_BINARY_SUFFIX2 = self.MODULE_BINARY_SUFFIX + '_2'
        self.MODULE_BINARY_CONTENTS = os.urandom(20)
        self.MODULE_BINARY_CONTENTS2 = b'\x00\xFF\xea\x9c\x11\xfeok\xb1\x8ax'

        self.module_name_order = [
            # 0
            {'suffix': self.MODULE_BINARY_SUFFIX,
             'priority': True, 'order': 1},
            # 1
            {'suffix': self.MODULE_BINARY_SUFFIX2,
             'priority': True, 'order': 2},
            # 2
            {'suffix': '_hidden_all_tenant_auto_priority',
             'priority': True, 'order': 3},
            # 3
            {'suffix': '_hidden', 'priority': True, 'order': 4},
            # 4
            {'suffix': '_auto', 'priority': True, 'order': 5},
            # 5
            {'suffix': '_live', 'priority': True, 'order': 6},
            # 6
            {'suffix': '_priority', 'priority': True, 'order': 7},
            # 7
            {'suffix': '_ds', 'priority': False, 'order': 1},
            # 8
            {'suffix': '_ds_ver', 'priority': False, 'order': 2},
            # 9
            {'suffix': '_all_tenant_ds_ver', 'priority': False, 'order': 3},
            # 10
            {'suffix': '', 'priority': False, 'order': 4},
            # 11
            {'suffix': '_ds_diff', 'priority': False, 'order': 5},
            # 12
            {'suffix': '_diff_tenant', 'priority': False, 'order': 6},
            # 13
            {'suffix': '_full_access', 'priority': False, 'order': 7},
            # 14
            {'suffix': '_for_update', 'priority': False, 'order': 8},
            # 15
            {'suffix': '_updated', 'priority': False, 'order': 8},
        ]

        self.apply_count = 0
        self.mod_inst_id = None
        self.mod_inst_apply_count = 0
        self.temp_module = None
        self.live_update_orig_md5 = None
        self.reapply_max_upd_date = None
        self._module_type = None

        self.test_modules = []
        self.module_count_prior_to_create = 0
        self.module_ds_count_prior_to_create = 0
        self.module_ds_all_count_prior_to_create = 0
        self.module_auto_apply_count_prior_to_create = 0
        self.module_admin_count_prior_to_create = 0
        self.module_other_count_prior_to_create = 0

        self.module_create_count = 0
        self.module_ds_create_count = 0
        self.module_ds_all_create_count = 0
        self.module_all_tenant_create_count = 0
        self.module_auto_apply_create_count = 0
        self.module_admin_create_count = 0
        self.module_other_create_count = 0

    @property
    def module_type(self):
        if not self._module_type:
            self._module_type = self.test_helper.get_valid_module_type()
        return self._module_type

    def _get_test_module(self, index):
        if not self.test_modules or len(self.test_modules) < (index + 1):
            raise SkipTest("Requested module not created")
        return self.test_modules[index]

    @property
    def main_test_module(self):
        # The module named "test_module_1"
        return self._get_test_module(0)

    @property
    def update_test_module(self):
        # The module named "test_module_1_updated"
        return self._get_test_module(1)

    @property
    def live_update_test_module(self):
        return self._find_live_update_module()

    def build_module_args(self, name_order=None):
        suffix = "_unknown"
        priority = False
        order = 5
        if name_order is not None:
            name_rec = self.module_name_order[name_order]
            suffix = name_rec['suffix']
            priority = name_rec['priority']
            order = name_rec['order']
        name = self.MODULE_NAME + suffix
        description = self.MODULE_DESC + suffix.replace('_', ' ')
        contents = self.get_module_contents(name)
        return name, description, contents, priority, order

    def get_module_contents(self, name=None):
        message = self.get_module_message(name=name)
        return self.MODULE_CONTENTS_PATTERN % message

    def get_module_message(self, name=None):
        name = name or self.MODULE_NAME
        return self.MODULE_MESSAGE_PATTERN % name

    def _find_invisible_module(self):
        def _match(mod):
            return not mod.visible and mod.tenant_id and not mod.auto_apply
        return self._find_module(_match, "Could not find invisible module")

    def _find_module(self, match_fn, not_found_message, find_all=False,
                     fail_on_not_found=True):
        found = [] if find_all else None
        for test_module in self.test_modules:
            if match_fn(test_module):
                if find_all:
                    found.append(test_module)
                else:
                    found = test_module
                    break
        if not found:
            if fail_on_not_found:
                self.fail(not_found_message)
            else:
                SkipTest(not_found_message)
        return found

    def _find_auto_apply_module(self):
        def _match(mod):
            return mod.auto_apply and mod.tenant_id and mod.visible
        return self._find_module(_match, "Could not find auto-apply module")

    def _find_live_update_module(self):
        def _match(mod):
            return mod.live_update and mod.tenant_id and mod.visible
        return self._find_module(_match, "Could not find live-update module")

    def _find_all_tenant_module(self):
        def _match(mod):
            return mod.tenant_id is None and mod.visible
        return self._find_module(_match, "Could not find all tenant module")

    def _find_priority_apply_module(self):
        def _match(mod):
            return mod.priority_apply and mod.tenant_id and mod.visible
        return self._find_module(_match,
                                 "Could not find priority-apply module")

    def _find_diff_datastore_module(self):
        def _match(mod):
            return (mod.datastore and
                    mod.datastore != models.Modules.MATCH_ALL_NAME and
                    mod.datastore != self.instance_info.dbaas_datastore)
        return self._find_module(_match,
                                 "Could not find different datastore module",
                                 fail_on_not_found=False)

    def _find_all_auto_apply_modules(self, visible=None):
        def _match(mod):
            return mod.auto_apply and (
                visible is None or mod.visible == visible)
        return self._find_module(
            _match, "Could not find all auto apply modules", find_all=True)

    def _find_module_by_id(self, module_id):
        def _match(mod):
            return mod.id == module_id
        return self._find_module(_match, "Could not find module with id %s" %
                                 module_id)

    # Tests start here
    def run_module_delete_existing(self):
        """Delete all the testing modules if exist."""
        modules = self.admin_client.modules.list()
        for module in modules:
            if module.name.startswith(self.MODULE_NAME):
                self.admin_client.modules.delete(module.id)

    def run_module_create_bad_type(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, 'invalid-type', self.MODULE_NEG_CONTENTS)

    def run_module_create_non_admin_auto(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        """Non-admin cannot create modules by specifying auto_apply."""
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, self.module_type, self.MODULE_NEG_CONTENTS,
            auto_apply=True)

    def run_module_create_non_admin_all_tenant(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        """Non-admin cannot create modules by specifying all_tenants."""
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, self.module_type, self.MODULE_NEG_CONTENTS,
            all_tenants=True)

    def run_module_create_non_admin_hidden(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        """Non-admin cannot create modules by specifying visible."""
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, self.module_type, self.MODULE_NEG_CONTENTS,
            visible=False)

    def run_module_create_non_admin_priority(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        """Non-admin cannot create modules by specifying priority_apply."""
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, self.module_type, self.MODULE_NEG_CONTENTS,
            priority_apply=True)

    def run_module_create_non_admin_no_full_access(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        """Non-admin cannot create modules by specifying full_access."""
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, self.module_type, self.MODULE_NEG_CONTENTS,
            full_access=False)

    def run_module_create_full_access_with_admin_opt(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        """full_access cannot be used together with auto_apply."""
        client = self.admin_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, self.module_type, self.MODULE_NEG_CONTENTS,
            full_access=True, auto_apply=True)

    def run_module_create_bad_datastore(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, self.module_type, self.MODULE_NEG_CONTENTS,
            datastore='bad-datastore')

    def run_module_create_bad_datastore_version(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, self.module_type, self.MODULE_NEG_CONTENTS,
            datastore=self.instance_info.dbaas_datastore,
            datastore_version='bad-datastore-version')

    def run_module_create_missing_datastore(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, self.module_type, self.MODULE_NEG_CONTENTS,
            datastore_version=self.instance_info.dbaas_datastore_version)

    def run_module_create(self):
        # Necessary to test that the count increases.
        self.module_count_prior_to_create = len(
            self.auth_client.modules.list())
        self.module_ds_count_prior_to_create = len(
            self.auth_client.modules.list(
                datastore=self.instance_info.dbaas_datastore))
        self.module_ds_all_count_prior_to_create = len(
            self.auth_client.modules.list(
                datastore=models.Modules.MATCH_ALL_NAME))
        self.module_auto_apply_count_prior_to_create = len(
            [module for module in self.admin_client.modules.list()
             if module.auto_apply])
        self.module_admin_count_prior_to_create = len(
            self.admin_client.modules.list())
        self.module_other_count_prior_to_create = len(
            self.unauth_client.modules.list())

        # Create module "test_module_1" for datastore "all"
        self.assert_module_create(self.auth_client, 10)

    def assert_module_create(self, client, name_order,
                             name=None, module_type=None,
                             contents=None, description=None,
                             all_tenants=False,
                             datastore=None, datastore_version=None,
                             auto_apply=False,
                             live_update=False, visible=True,
                             priority_apply=None,
                             apply_order=None,
                             full_access=None):
        (temp_name, temp_description, temp_contents,
         temp_priority, temp_order) = self.build_module_args(name_order)
        name = name if name is not None else temp_name
        description = (
            description if description is not None else temp_description)
        contents = contents if contents is not None else temp_contents
        priority_apply = (
            priority_apply if priority_apply is not None else temp_priority)
        apply_order = apply_order if apply_order is not None else temp_order
        module_type = module_type or self.module_type
        result = client.modules.create(
            name, module_type, contents,
            description=description,
            all_tenants=all_tenants,
            datastore=datastore, datastore_version=datastore_version,
            auto_apply=auto_apply,
            live_update=live_update, visible=visible,
            priority_apply=priority_apply,
            apply_order=apply_order,
            full_access=full_access)

        username = client.real_client.client.username
        if username == self.instance_info.user.auth_user:
            self.module_create_count += 1
            if datastore:
                if datastore == self.instance_info.dbaas_datastore:
                    self.module_ds_create_count += 1
            else:
                self.module_ds_all_create_count += 1
        elif (username != self.instance_info.admin_user.auth_user and
              username != self.instance_info.user.auth_user):
            self.module_other_create_count += 1
        else:
            self.module_admin_create_count += 1

        if all_tenants and visible:
            self.module_all_tenant_create_count += 1
            if datastore:
                if datastore == self.instance_info.dbaas_datastore:
                    self.module_ds_create_count += 1
            else:
                self.module_ds_all_create_count += 1
        if auto_apply and visible:
            self.module_auto_apply_create_count += 1

        self.test_modules.append(result)

        tenant_id = None
        tenant = models.Modules.MATCH_ALL_NAME
        if not all_tenants:
            tenant, tenant_id = self.get_client_tenant(client)
            # If we find a way to grab the tenant name in the module
            # stuff, the line below can be removed
            tenant = tenant_id
        datastore = datastore or models.Modules.MATCH_ALL_NAME
        datastore_version = datastore_version or models.Modules.MATCH_ALL_NAME
        self.validate_module(
            result, validate_all=False,
            expected_name=name,
            expected_module_type=module_type,
            expected_description=description,
            expected_tenant=tenant,
            expected_tenant_id=tenant_id,
            expected_datastore=datastore,
            expected_datastore_version=datastore_version,
            expected_auto_apply=auto_apply,
            expected_contents=contents,
            expected_is_admin=(
                username == self.instance_info.admin_user.auth_user and
                not full_access
            )
        )

    def validate_module(self, module, validate_all=False,
                        expected_name=None,
                        expected_module_type=None,
                        expected_description=None,
                        expected_tenant=None,
                        expected_tenant_id=None,
                        expected_datastore=None,
                        expected_datastore_id=None,
                        expected_all_datastores=None,
                        expected_datastore_version=None,
                        expected_datastore_version_id=None,
                        expected_all_datastore_versions=None,
                        expected_all_tenants=None,
                        expected_auto_apply=None,
                        expected_live_update=None,
                        expected_visible=None,
                        expected_contents=None,
                        expected_priority_apply=None,
                        expected_apply_order=None,
                        expected_is_admin=None,
                        expected_full_access=None):

        if expected_all_tenants:
            expected_tenant = expected_tenant or models.Modules.MATCH_ALL_NAME
        if expected_all_datastores:
            expected_datastore = models.Modules.MATCH_ALL_NAME
            expected_datastore_id = None
        if expected_all_datastore_versions:
            expected_datastore_version = models.Modules.MATCH_ALL_NAME
            expected_datastore_version_id = None
        if expected_name:
            self.assert_equal(expected_name, module.name,
                              'Unexpected module name')
        if expected_module_type:
            self.assert_equal(expected_module_type.lower(), module.type,
                              'Unexpected module type')
        if expected_description:
            self.assert_equal(expected_description, module.description,
                              'Unexpected module description')
        if expected_tenant_id:
            self.assert_equal(expected_tenant_id, module.tenant_id,
                              'Unexpected tenant id')
        if expected_tenant:
            self.assert_equal(expected_tenant, module.tenant,
                              'Unexpected tenant name')
        if expected_datastore:
            self.assert_equal(expected_datastore, module.datastore,
                              'Unexpected datastore')
        if expected_datastore_version:
            self.assert_equal(expected_datastore_version,
                              module.datastore_version,
                              'Unexpected datastore version')
        if expected_auto_apply is not None:
            self.assert_equal(expected_auto_apply, module.auto_apply,
                              'Unexpected auto_apply')
        if expected_priority_apply is not None:
            self.assert_equal(expected_priority_apply, module.priority_apply,
                              'Unexpected priority_apply')
        if expected_apply_order is not None:
            self.assert_equal(expected_apply_order, module.apply_order,
                              'Unexpected apply_order')
        if expected_is_admin is not None:
            self.assert_equal(expected_is_admin, module.is_admin,
                              'Unexpected is_admin')
        if expected_full_access is not None:
            self.assert_equal(expected_full_access, not module.is_admin,
                              'Unexpected full_access')
        if validate_all:
            if expected_datastore_id:
                self.assert_equal(expected_datastore_id, module.datastore_id,
                                  'Unexpected datastore id')
            if expected_datastore_version_id:
                self.assert_equal(expected_datastore_version_id,
                                  module.datastore_version_id,
                                  'Unexpected datastore version id')
            if expected_live_update is not None:
                self.assert_equal(expected_live_update, module.live_update,
                                  'Unexpected live_update')
            if expected_visible is not None:
                self.assert_equal(expected_visible, module.visible,
                                  'Unexpected visible')

    def run_module_create_for_update(self):
        # Create module "test_module_1_updated"
        self.assert_module_create(self.auth_client, 14)

    def run_module_create_dupe(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.create,
            self.MODULE_NAME, self.module_type, self.MODULE_NEG_CONTENTS)

    def run_module_update_missing_datastore(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update,
            self.update_test_module.id,
            datastore_version=self.instance_info.dbaas_datastore_version)

    def run_module_create_bin(self):
        self.assert_module_create(
            self.admin_client, 0,
            contents=self.MODULE_BINARY_CONTENTS,
            auto_apply=True, visible=False)

    def run_module_create_bin2(self):
        self.assert_module_create(
            self.admin_client, 1,
            contents=self.MODULE_BINARY_CONTENTS2,
            auto_apply=True, visible=False)

    def run_module_show(self):
        test_module = self.main_test_module
        result = self.auth_client.modules.get(test_module.id)
        self.validate_module(
            result, validate_all=True,
            expected_name=test_module.name,
            expected_module_type=test_module.type,
            expected_description=test_module.description,
            expected_tenant=test_module.tenant,
            expected_datastore=test_module.datastore,
            expected_datastore_version=test_module.datastore_version,
            expected_auto_apply=test_module.auto_apply,
            expected_live_update=False,
            expected_visible=True,
            expected_priority_apply=test_module.priority_apply,
            expected_apply_order=test_module.apply_order,
            expected_is_admin=test_module.is_admin)

    def run_module_show_unauth_user(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        client = self.unauth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.get, self.main_test_module.id)

    def run_module_list(self):
        self.assert_module_list(
            self.auth_client,
            (
                self.module_count_prior_to_create +
                self.module_create_count +
                self.module_all_tenant_create_count
            )
        )

    def assert_module_list(self, client, expected_count, datastore=None):
        if datastore:
            module_list = client.modules.list(datastore=datastore)
        else:
            module_list = client.modules.list()
        self.assert_equal(expected_count, len(module_list),
                          "Wrong number of modules for list")
        for module in module_list:
            # only validate the test modules
            if module.name.startswith(self.MODULE_NAME):
                test_module = self._find_module_by_id(module.id)
                self.validate_module(
                    module, validate_all=True,
                    expected_name=test_module.name,
                    expected_module_type=test_module.type,
                    expected_description=test_module.description,
                    expected_tenant=test_module.tenant,
                    expected_datastore=test_module.datastore,
                    expected_datastore_version=test_module.datastore_version,
                    expected_auto_apply=test_module.auto_apply,
                    expected_priority_apply=test_module.priority_apply,
                    expected_apply_order=test_module.apply_order,
                    expected_is_admin=test_module.is_admin)

    def run_module_list_unauth_user(self):
        self.assert_module_list(
            self.unauth_client,
            (self.module_other_count_prior_to_create +
             self.module_all_tenant_create_count +
             self.module_other_create_count))

    def run_module_create_admin_all(self):
        self.assert_module_create(
            self.admin_client, 2,
            all_tenants=True,
            visible=False,
            auto_apply=True)

    def run_module_create_admin_hidden(self):
        self.assert_module_create(
            self.admin_client, 3,
            visible=False)

    def run_module_create_admin_auto(self):
        self.assert_module_create(
            self.admin_client, 4,
            auto_apply=True)

    def run_module_create_admin_live_update(self):
        self.assert_module_create(
            self.admin_client, 5,
            live_update=True)
        self.live_update_orig_md5 = self.test_modules[-1].md5

    def run_module_create_admin_priority_apply(self):
        self.assert_module_create(
            self.admin_client, 6)

    def run_module_create_datastore(self):
        self.assert_module_create(
            self.admin_client, 7,
            datastore=self.instance_info.dbaas_datastore)

    def run_module_create_different_datastore(self):
        diff_datastore = self._get_different_datastore()
        if not diff_datastore:
            raise SkipTest("Could not find a different datastore")
        self.assert_module_create(
            self.auth_client, 11,
            datastore=diff_datastore)

    def _get_different_datastore(self):
        different_datastore = None
        datastores = self.admin_client.datastores.list()
        for datastore in datastores:
            self.report.log("Found datastore: %s" % datastore.name)
            if datastore.name != self.instance_info.dbaas_datastore:
                different_datastore = datastore.name
                break
        return different_datastore

    def run_module_create_ds_version(self):
        self.assert_module_create(
            self.admin_client, 8,
            datastore=self.instance_info.dbaas_datastore,
            datastore_version=self.instance_info.dbaas_datastore_version)

    def run_module_create_all_tenant(self):
        self.assert_module_create(
            self.admin_client, 9,
            all_tenants=True,
            datastore=self.instance_info.dbaas_datastore,
            datastore_version=self.instance_info.dbaas_datastore_version)

    def run_module_create_different_tenant(self):
        self.assert_module_create(
            self.unauth_client, 12)

    def run_module_create_full_access(self):
        self.assert_module_create(
            self.admin_client, 13,
            full_access=True)

    def run_module_full_access_toggle(self):
        self.assert_module_update(
            self.admin_client,
            self.main_test_module.id,
            full_access=False)
        self.assert_module_update(
            self.admin_client,
            self.main_test_module.id,
            full_access=True)

    def run_module_list_again(self):
        self.assert_module_list(
            self.auth_client,
            (
                self.module_count_prior_to_create +
                self.module_create_count +
                self.module_all_tenant_create_count
            )
        )

    def run_module_list_ds(self):
        self.assert_module_list(
            self.auth_client,
            self.module_ds_count_prior_to_create + self.module_ds_create_count,
            datastore=self.instance_info.dbaas_datastore)

    def run_module_list_ds_all(self):
        self.assert_module_list(
            self.auth_client,
            (self.module_ds_all_count_prior_to_create +
             self.module_ds_all_create_count),
            datastore=models.Modules.MATCH_ALL_NAME)

    def run_module_show_invisible(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        module = self._find_invisible_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.get, module.id)

    def run_module_list_admin(self):
        self.assert_module_list(
            self.admin_client,
            (self.module_admin_count_prior_to_create +
             self.module_create_count +
             self.module_admin_create_count +
             self.module_other_create_count))

    def run_module_update(self):
        self.assert_module_update_description(
            self.auth_client,
            self.main_test_module.id,
            description=self.MODULE_DESC + " modified")

    def assert_module_update(self, client, module_id, **kwargs):
        result = client.modules.update(module_id, **kwargs)
        found = False
        index = -1
        for test_module in self.test_modules:
            index += 1
            if test_module.id == module_id:
                found = True
                break
        if not found:
            self.fail("Could not find updated module in module list")
        self.test_modules[index] = result

        expected_args = {}
        for key, value in kwargs.items():
            new_key = 'expected_' + key
            expected_args[new_key] = value
        self.validate_module(result, **expected_args)

    def run_module_update_same_contents(self):
        old_md5 = self.main_test_module.md5
        self.assert_module_update(
            self.auth_client,
            self.main_test_module.id,
            contents=self.get_module_contents(self.main_test_module.name))
        self.assert_equal(old_md5, self.main_test_module.md5,
                          "MD5 changed with same contents")

    def run_module_update_auto_toggle(self,
                                      expected_exception=exceptions.NotFound,
                                      expected_http_code=404):
        module = self._find_auto_apply_module()
        toggle_off_args = {'auto_apply': False}
        toggle_on_args = {'auto_apply': True}
        self.assert_module_toggle(module, toggle_off_args, toggle_on_args,
                                  expected_exception=expected_exception,
                                  expected_http_code=expected_http_code)

    def assert_module_toggle(self, module, toggle_off_args, toggle_on_args,
                             expected_exception, expected_http_code):
        # First try to update the module based on the change
        # (this should toggle the state but still not allow non-admin access)
        client = self.admin_client
        self.assert_module_update(client, module.id, **toggle_off_args)
        # The non-admin client should fail to update
        non_admin_client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            non_admin_client, non_admin_client.modules.update, module.id,
            description='Updated by non-admin')
        # Make sure we can still update with the admin client
        self.assert_module_update(
            client, module.id, description='Updated by admin')
        # Now set it back
        self.assert_module_update(
            client, module.id, description=module.description,
            **toggle_on_args)

    def run_module_update_all_tenant_toggle(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        module = self._find_all_tenant_module()
        toggle_off_args = {'all_tenants': False}
        toggle_on_args = {'all_tenants': True}
        self.assert_module_toggle(module, toggle_off_args, toggle_on_args,
                                  expected_exception=expected_exception,
                                  expected_http_code=expected_http_code)

    def run_module_update_invisible_toggle(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        module = self._find_invisible_module()
        toggle_off_args = {'visible': True}
        toggle_on_args = {'visible': False}
        self.assert_module_toggle(module, toggle_off_args, toggle_on_args,
                                  expected_exception=expected_exception,
                                  expected_http_code=expected_http_code)

    def run_module_update_priority_toggle(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        module = self._find_priority_apply_module()
        toggle_off_args = {'priority_apply': False}
        toggle_on_args = {'priority_apply': True}
        self.assert_module_toggle(module, toggle_off_args, toggle_on_args,
                                  expected_exception=expected_exception,
                                  expected_http_code=expected_http_code)

    def run_module_update_unauth(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        client = self.unauth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update,
            self.main_test_module.id, description='Upd')

    def run_module_update_non_admin_auto(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update,
            self.main_test_module.id, visible=False)

    def run_module_update_non_admin_auto_off(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        module = self._find_auto_apply_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update, module.id, auto_apply=False)

    def run_module_update_non_admin_auto_any(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        module = self._find_auto_apply_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update, module.id, description='Upd')

    def run_module_update_non_admin_all_tenant(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update,
            self.main_test_module.id, all_tenants=True)

    def run_module_update_non_admin_all_tenant_off(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        module = self._find_all_tenant_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update, module.id, all_tenants=False)

    def run_module_update_non_admin_all_tenant_any(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        module = self._find_all_tenant_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update, module.id, description='Upd')

    def run_module_update_non_admin_invisible(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update,
            self.main_test_module.id, visible=False)

    def run_module_update_non_admin_invisible_off(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        module = self._find_invisible_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update, module.id, visible=True)

    def run_module_update_non_admin_invisible_any(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        module = self._find_invisible_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update, module.id, description='Upd')

    # ModuleInstanceGroup methods
    def run_module_list_instance_empty(self):
        self.assert_module_list_instance(
            self.auth_client, self.instance_info.id,
            self.module_auto_apply_count_prior_to_create)

    def assert_module_list_instance(self, client, instance_id, expected_count,
                                    expected_http_code=200):
        module_list = client.instances.modules(instance_id)
        self.assert_client_code(client, expected_http_code)
        count = len(module_list)
        self.assert_equal(expected_count, count,
                          "Wrong number of modules from list instance")

        for module in module_list:
            self.validate_module(module)

    def run_module_instances_empty(self):
        self.assert_module_instances(
            self.auth_client, self.main_test_module.id, 0)

    def assert_module_instances(self, client, module_id, expected_count,
                                expected_http_code=200):
        instance_list = client.modules.instances(module_id)
        self.assert_client_code(client, expected_http_code)
        count = len(instance_list)
        self.assert_equal(expected_count, count,
                          "Wrong number of instances applied from module")

    def run_module_instance_count_empty(self):
        self.assert_module_instance_count(
            self.auth_client, self.main_test_module.id, 0)

    def assert_module_instance_count(self, client, module_id,
                                     expected_rows, expected_count=None,
                                     expected_http_code=200):
        instance_count_list = client.modules.instances(module_id,
                                                       count_only=True)
        self.assert_client_code(client, expected_http_code)
        rowcount = len(instance_count_list)
        self.assert_equal(expected_rows, rowcount,
                          "Wrong number of instance count records from module")
        # expected_count is a dict of md5->count pairs.
        if expected_rows and expected_count:
            for row in instance_count_list:
                self.assert_equal(
                    expected_count[row.module_md5], row.instance_count,
                    "Wrong count in record from module instances; md5: %s" %
                    row.module_md5)

    def run_module_query_empty(self):
        self.assert_module_query(
            self.auth_client, self.instance_info.id,
            self.module_auto_apply_count_prior_to_create)

    def assert_module_query(self, client, instance_id, expected_count,
                            expected_http_code=200, expected_results=None):
        modquery_list = client.instances.module_query(instance_id)
        self.assert_client_code(client, expected_http_code)
        count = len(modquery_list)
        self.assert_equal(expected_count, count,
                          "Wrong number of modules from query")
        expected_results = expected_results or {}
        name_index = len(self.module_name_order)
        for modquery in modquery_list:
            if modquery.name in expected_results:
                self.report.log("Validating module '%s'" % modquery.name)
                expected = expected_results[modquery.name]
                self.validate_module_apply_info(
                    modquery,
                    expected_status=expected['status'],
                    expected_message=expected['message'])
                # make sure we're in the correct order
                found = False
                while name_index > 0:
                    name_index -= 1
                    name_order_rec = self.module_name_order[name_index]
                    order_name = self.MODULE_NAME + name_order_rec['suffix']
                    self.report.log("Next module order '%s'" % order_name)
                    if order_name == modquery.name:
                        self.report.log("Match found")
                        found = True
                        break
                if name_index == 0 and not found:
                    self.fail("Module '%s' was not found in the correct order"
                              % modquery.name)

    def run_module_apply(self):
        self.assert_module_apply(self.auth_client, self.instance_info.id,
                                 self.main_test_module)
        self.apply_count += 1

    def assert_module_apply(self, client, instance_id, module,
                            expected_is_admin=False,
                            expected_status=None, expected_message=None,
                            expected_contents=None,
                            expected_http_code=200):
        module_apply_list = client.instances.module_apply(
            instance_id, [module.id])
        self.assert_client_code(client, expected_http_code)
        expected_status = expected_status or 'OK'
        expected_message = (expected_message or
                            self.get_module_message(module.name))
        for module_apply in module_apply_list:
            self.validate_module_apply_info(
                module_apply,
                expected_name=module.name,
                expected_module_type=module.type,
                expected_datastore=module.datastore,
                expected_datastore_version=module.datastore_version,
                expected_auto_apply=module.auto_apply,
                expected_visible=module.visible,
                expected_contents=expected_contents,
                expected_status=expected_status,
                expected_message=expected_message,
                expected_is_admin=expected_is_admin)

    def validate_module_apply_info(self, module_apply,
                                   expected_name=None,
                                   expected_module_type=None,
                                   expected_datastore=None,
                                   expected_datastore_version=None,
                                   expected_auto_apply=None,
                                   expected_visible=None,
                                   expected_contents=None,
                                   expected_message=None,
                                   expected_status=None,
                                   expected_is_admin=None):

        prefix = "Module: %s -" % expected_name
        if expected_name:
            self.assert_equal(expected_name, module_apply.name,
                              '%s Unexpected module name' % prefix)
        if expected_module_type:
            self.assert_equal(expected_module_type, module_apply.type,
                              '%s Unexpected module type' % prefix)
        if expected_datastore:
            self.assert_equal(expected_datastore, module_apply.datastore,
                              '%s Unexpected datastore' % prefix)
        if expected_datastore_version:
            self.assert_equal(expected_datastore_version,
                              module_apply.datastore_version,
                              '%s Unexpected datastore version' % prefix)
        if expected_auto_apply is not None:
            self.assert_equal(expected_auto_apply, module_apply.auto_apply,
                              '%s Unexpected auto_apply' % prefix)
        if expected_visible is not None:
            self.assert_equal(expected_visible, module_apply.visible,
                              '%s Unexpected visible' % prefix)
        if expected_contents is not None:
            self.assert_equal(expected_contents, module_apply.contents,
                              '%s Unexpected contents' % prefix)
        if expected_message is not None:
            regex = re.compile(expected_message)
            self.assert_true(regex.match(module_apply.message),
                             "%s Unexpected message '%s', expected '%s'" %
                             (prefix, module_apply.message, expected_message))
        if expected_status is not None:
            self.assert_equal(expected_status, module_apply.status,
                              '%s Unexpected status' % prefix)
        if expected_is_admin is not None:
            self.assert_equal(expected_is_admin, module_apply.is_admin,
                              '%s Unexpected is_admin' % prefix)

    def run_module_apply_wrong_module(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        module = self._find_diff_datastore_module()
        self.report.log("Found 'wrong' module: %s" % module.name)
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.instances.module_apply,
            self.instance_info.id, [module.id])

    def run_module_list_instance_after_apply(self):
        self.assert_module_list_instance(
            self.auth_client,
            self.instance_info.id,
            self.apply_count + self.module_auto_apply_count_prior_to_create
        )

    def run_module_apply_another(self):
        self.assert_module_apply(self.auth_client, self.instance_info.id,
                                 self.update_test_module)
        self.apply_count += 1

    def run_module_list_instance_after_apply_another(self):
        self.assert_module_list_instance(
            self.auth_client,
            self.instance_info.id,
            self.apply_count + self.module_auto_apply_count_prior_to_create
        )

    def run_module_update_after_remove(self):
        name, description, contents, priority, order = (
            self.build_module_args(15))
        self.assert_module_update(
            self.auth_client,
            self.update_test_module.id,
            name=name,
            datastore=self.instance_info.dbaas_datastore,
            datastore_version=self.instance_info.dbaas_datastore_version,
            contents=contents)

    def run_module_instances_after_apply(self):
        self.assert_module_instances(
            self.auth_client, self.main_test_module.id, 1)

    def run_module_instance_count_after_apply(self):
        self.assert_module_instance_count(
            self.auth_client, self.main_test_module.id, 1,
            {self.main_test_module.md5: 1})

    def run_module_query_after_apply(self):
        expected_count = (self.module_auto_apply_count_prior_to_create +
                          self.apply_count)
        expected_results = self.create_default_query_expected_results(
            [self.main_test_module])
        self.assert_module_query(self.auth_client, self.instance_info.id,
                                 expected_count=expected_count,
                                 expected_results=expected_results)

    def create_default_query_expected_results(self, modules, is_admin=False):
        expected_results = {}
        for module in modules:
            status = 'OK'
            message = self.get_module_message(module.name)
            contents = self.get_module_contents(module.name)
            if not is_admin and (not module.visible or module.auto_apply or
                                 not module.tenant_id):
                contents = ('Must be admin to retrieve contents for module %s'
                            % module.name)
            elif self.MODULE_BINARY_SUFFIX in module.name:
                status = 'ERROR'
                message = ('^(Could not extract ping message|'
                           'Message not found in contents file).*')
                contents = self.MODULE_BINARY_CONTENTS
                if self.MODULE_BINARY_SUFFIX2 in module.name:
                    contents = self.MODULE_BINARY_CONTENTS2
            expected_results[module.name] = {
                'status': status,
                'message': message,
                'datastore': module.datastore,
                'datastore_version': module.datastore_version,
                'contents': contents,
            }
        return expected_results

    def run_module_instances_after_apply_another(self):
        self.assert_module_instances(
            self.auth_client, self.main_test_module.id, 1)

    def run_module_instance_count_after_apply_another(self):
        self.assert_module_instance_count(
            self.auth_client, self.main_test_module.id, 1,
            {self.main_test_module.md5: 1})

    def run_module_query_after_apply_another(self):
        expected_count = (self.module_auto_apply_count_prior_to_create +
                          self.apply_count)
        expected_results = self.create_default_query_expected_results(
            [self.main_test_module, self.update_test_module])
        self.assert_module_query(self.auth_client, self.instance_info.id,
                                 expected_count=expected_count,
                                 expected_results=expected_results)

    def run_module_update_not_live(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.update,
            self.main_test_module.id, description='Do not allow this change')

    def run_module_apply_live_update(self):
        module = self.live_update_test_module
        self.assert_module_apply(self.auth_client, self.instance_info.id,
                                 module, expected_is_admin=module.is_admin)
        self.apply_count += 1

    def run_module_list_instance_after_apply_live(self):
        self.assert_module_list_instance(
            self.auth_client,
            self.instance_info.id,
            self.apply_count + self.module_auto_apply_count_prior_to_create
        )

    def run_module_update_live_update(self):
        module = self.live_update_test_module
        new_contents = self.get_module_contents(name=module.name + '_upd')
        self.assert_module_update(
            self.admin_client,
            module.id,
            contents=new_contents)

    def run_module_update_after_remove_again(self):
        self.assert_module_update(
            self.auth_client,
            self.update_test_module.id,
            name=self.MODULE_NAME + '_updated_back',
            all_datastores=True,
            all_datastore_versions=True)

    def run_create_inst_with_mods(self, expected_http_code=200):
        live_update = self.live_update_test_module
        self.mod_inst_id = self.assert_inst_mod_create(
            [self.main_test_module.id, live_update.id],
            '_module', expected_http_code)
        self.mod_inst_apply_count += 2

    def assert_inst_mod_create(self, module_ids, name_suffix,
                               expected_http_code):
        client = self.auth_client
        inst = client.instances.create(
            self.instance_info.name + name_suffix,
            self.instance_info.dbaas_flavor_href,
            self.instance_info.volume,
            datastore=self.instance_info.dbaas_datastore,
            datastore_version=self.instance_info.dbaas_datastore_version,
            nics=self.instance_info.nics,
            modules=module_ids,
        )
        self.assert_client_code(client, expected_http_code)
        self.register_debug_inst_ids(inst.id)
        return inst.id

    def run_create_inst_with_wrong_module(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        module = self._find_diff_datastore_module()
        self.report.log("Found 'wrong' module: %s" % module.name)

        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.instances.create,
            self.instance_info.name + '_wrong_ds',
            self.instance_info.dbaas_flavor_href,
            self.instance_info.volume,
            datastore=self.instance_info.dbaas_datastore,
            datastore_version=self.instance_info.dbaas_datastore_version,
            nics=self.instance_info.nics,
            modules=[module.id])

    def run_module_delete_applied(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.delete, self.main_test_module.id)

    def run_module_remove(self):
        self.assert_module_remove(self.auth_client, self.instance_info.id,
                                  self.update_test_module.id)
        self.apply_count -= 1

    def assert_module_remove(self, client, instance_id, module_id,
                             expected_http_code=200):
        client.instances.module_remove(instance_id, module_id)
        self.assert_client_code(client, expected_http_code)

    def run_wait_for_inst_with_mods(self,
                                    expected_states=['BUILD', 'HEALTHY']):
        self.assert_instance_action(self.mod_inst_id, expected_states)

    def run_module_query_after_inst_create(self):
        auto_modules = self._find_all_auto_apply_modules(visible=True)
        expected_count = self.mod_inst_apply_count + len(auto_modules)
        expected_results = self.create_default_query_expected_results(
            [self.main_test_module] + auto_modules)
        self.assert_module_query(self.auth_client, self.mod_inst_id,
                                 expected_count=expected_count,
                                 expected_results=expected_results)

    def run_module_retrieve_after_inst_create(self):
        auto_modules = self._find_all_auto_apply_modules(visible=True)
        expected_count = self.mod_inst_apply_count + len(auto_modules)
        expected_results = self.create_default_query_expected_results(
            [self.main_test_module] + auto_modules)
        self.assert_module_retrieve(self.auth_client, self.mod_inst_id,
                                    expected_count=expected_count,
                                    expected_results=expected_results)

    def assert_module_retrieve(self, client, instance_id, expected_count,
                               expected_http_code=200, expected_results=None):
        try:
            temp_dir = tempfile.mkdtemp()
            prefix = 'contents'
            modretrieve_list = client.instances.module_retrieve(
                instance_id, directory=temp_dir, prefix=prefix)
            self.assert_client_code(client, expected_http_code)
            count = len(modretrieve_list)
            self.assert_equal(expected_count, count,
                              "Wrong number of modules from retrieve")
            expected_results = expected_results or {}
            for module_name, filename in modretrieve_list.items():
                if module_name in expected_results:
                    expected = expected_results[module_name]
                    contents_name = '%s_%s_%s_%s' % (
                        prefix, module_name,
                        expected['datastore'], expected['datastore_version'])
                    expected_filename = guestagent_utils.build_file_path(
                        temp_dir, contents_name, 'dat')
                    self.assert_equal(expected_filename, filename,
                                      'Unexpected retrieve filename')
                    if 'contents' in expected and expected['contents']:
                        with open(filename, 'rb') as fh:
                            contents = fh.read()

                        expected = expected['contents']
                        if isinstance(expected, str):
                            expected = expected.encode()

                        self.assert_equal(expected, contents,
                                          "Unexpected contents for %s" %
                                          module_name)
        finally:
            operating_system.remove(temp_dir)

    def run_module_query_after_inst_create_admin(self):
        auto_modules = self._find_all_auto_apply_modules()
        expected_count = self.mod_inst_apply_count + len(auto_modules)
        expected_results = self.create_default_query_expected_results(
            [self.main_test_module] + auto_modules, is_admin=True)
        self.assert_module_query(self.admin_client, self.mod_inst_id,
                                 expected_count=expected_count,
                                 expected_results=expected_results)

    def run_module_retrieve_after_inst_create_admin(self):
        auto_modules = self._find_all_auto_apply_modules()
        expected_count = self.mod_inst_apply_count + len(auto_modules)
        expected_results = self.create_default_query_expected_results(
            [self.main_test_module] + auto_modules, is_admin=True)
        self.assert_module_retrieve(self.admin_client, self.mod_inst_id,
                                    expected_count=expected_count,
                                    expected_results=expected_results)

    def run_module_delete_auto_applied(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        module = self._find_auto_apply_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.delete, module.id)

    def run_module_list_instance_after_mod_inst(self):
        self.assert_module_list_instance(
            self.auth_client, self.mod_inst_id,
            self.module_auto_apply_create_count + 2)

    def run_module_instances_after_mod_inst(self):
        self.assert_module_instances(
            self.auth_client, self.live_update_test_module.id, 2)

    def run_module_instance_count_after_mod_inst(self):
        self.assert_module_instance_count(
            self.auth_client, self.live_update_test_module.id, 2,
            {self.live_update_test_module.md5: 1,
             self.live_update_orig_md5: 1})

    def run_module_reapply_with_md5(self, expected_http_code=202):
        self.assert_module_reapply(
            self.auth_client, self.live_update_test_module,
            expected_http_code=expected_http_code,
            md5=self.live_update_test_module.md5)

    def assert_module_reapply(self, client, module, expected_http_code,
                              md5=None, force=False):
        self.reapply_max_upd_date = self.get_updated(client, module.id)
        client.modules.reapply(module.id, md5=md5, force=force)
        self.assert_client_code(client, expected_http_code)

    def run_module_reapply_with_md5_verify(self):
        # since this isn't supposed to do anything, we can't 'wait' for it to
        # finish, since we'll never know. So just sleep for a couple seconds
        # just to make sure.
        time.sleep(2)
        # Now we check that the max_updated_date field didn't change
        module_id = self.live_update_test_module.id
        instance_count_list = self.auth_client.modules.instances(
            module_id, count_only=True)
        mismatch = False
        for instance_count in instance_count_list:
            if self.reapply_max_upd_date != instance_count.max_updated_date:
                mismatch = True
        self.assert_true(
            mismatch,
            "Could not find record having max_updated_date different from %s" %
            self.reapply_max_upd_date)

    def run_module_list_instance_after_reapply_md5(self):
        self.assert_module_list_instance(
            self.auth_client, self.mod_inst_id,
            self.module_auto_apply_create_count + 2)

    def run_module_instances_after_reapply_md5(self):
        self.assert_module_instances(
            self.auth_client, self.live_update_test_module.id, 2)

    def run_module_instance_count_after_reapply_md5(self):
        self.assert_module_instance_count(
            self.auth_client, self.live_update_test_module.id, 2,
            {self.live_update_test_module.md5: 1,
             self.live_update_orig_md5: 1})

    def run_module_reapply_all(self, expected_http_code=202):
        module_id = self.live_update_test_module.id
        client = self.auth_client
        self.reapply_max_upd_date = self.get_updated(client, module_id)
        self.assert_module_reapply(
            client, self.live_update_test_module,
            expected_http_code=expected_http_code)

    def run_module_reapply_all_wait(self):
        self.wait_for_reapply(
            self.auth_client, self.live_update_test_module.id,
            md5=self.live_update_orig_md5)

    def wait_for_reapply(self, client, module_id, updated=None, md5=None):
        """Reapply is done when all the counts for 'md5' are gone.  If updated
        is passed in, the min_updated_date must all be greater than it.
        """
        if not updated and not md5:
            raise RuntimeError("Code error: Must pass in 'updated' or 'md5'.")
        self.report.log("Waiting for all md5:%s modules to have an updated "
                        "date greater than %s" % (md5, updated))

        def _all_updated():
            min_updated = self.get_updated(
                client, module_id, max=False, md5=md5)
            if md5:
                return min_updated is None
            return min_updated > updated

        timeout = 60
        try:
            poll_until(_all_updated, time_out=timeout, sleep_time=5)
            self.report.log("All instances now have the current module "
                            "for md5: %s." % md5)
        except exception.PollTimeOut:
            self.fail("Some instances were not updated with the "
                      "timeout: %ds" % timeout)

    def get_updated(self, client, module_id, max=True, md5=None):
        updated = None
        instance_count_list = client.modules.instances(
            module_id, count_only=True)
        for instance_count in instance_count_list:
            if not md5 or md5 == instance_count.module_md5:
                if not updated or (
                        (max and instance_count.max_updated_date > updated) or
                        (not max and
                         instance_count.min_updated_date < updated)):
                    updated = (instance_count.max_updated_date
                               if max else instance_count.min_updated_date)
        return updated

    def run_module_list_instance_after_reapply(self):
        self.assert_module_list_instance(
            self.auth_client, self.mod_inst_id,
            self.module_auto_apply_create_count + 2)

    def run_module_instances_after_reapply(self):
        self.assert_module_instances(
            self.auth_client, self.live_update_test_module.id, 2)

    def run_module_instance_count_after_reapply(self):
        self.assert_module_instance_count(
            self.auth_client, self.live_update_test_module.id, 1,
            {self.live_update_test_module.md5: 2})

    def run_module_reapply_with_force(self, expected_http_code=202):
        self.assert_module_reapply(
            self.auth_client, self.live_update_test_module,
            expected_http_code=expected_http_code,
            force=True)

    def run_module_reapply_with_force_wait(self):
        self.wait_for_reapply(
            self.auth_client, self.live_update_test_module.id,
            updated=self.reapply_max_upd_date)

    def run_delete_inst_with_mods(self, expected_http_code=202):
        self.assert_delete_instance(self.mod_inst_id, expected_http_code)

    def assert_delete_instance(self, instance_id, expected_http_code):
        client = self.auth_client
        client.instances.delete(instance_id)
        self.assert_client_code(client, expected_http_code)

    def run_remove_mods_from_main_inst(self, expected_http_code=200):
        client = self.auth_client
        modquery_list = client.instances.module_query(self.instance_info.id)
        self.assert_client_code(client, expected_http_code)
        for modquery in modquery_list:
            client.instances.module_remove(self.instance_info.id, modquery.id)
            self.assert_client_code(client, expected_http_code)

    def run_wait_for_delete_inst_with_mods(
            self, expected_last_state=['SHUTDOWN']):
        self.assert_all_gone(self.mod_inst_id, expected_last_state)

    # ModuleDeleteGroup methods
    def run_module_delete_non_existent(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.delete, 'bad_id')

    def run_module_delete_unauth_user(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        client = self.unauth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.delete, self.main_test_module.id)

    def run_module_delete_hidden_by_non_admin(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        module = self._find_invisible_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.delete, module.id)

    def run_module_delete_all_tenant_by_non_admin(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        module = self._find_all_tenant_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.delete, module.id)

    def run_module_delete_auto_by_non_admin(
            self, expected_exception=exceptions.Forbidden,
            expected_http_code=403):
        module = self._find_auto_apply_module()
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.modules.delete, module.id)

    def run_module_delete(self):
        expected_count = len(self.auth_client.modules.list()) - 1
        test_module = self.test_modules.pop(0)
        self.assert_module_delete(self.auth_client, test_module.id,
                                  expected_count)

    def run_module_delete_admin(self):
        start_count = count = len(self.admin_client.modules.list())
        for test_module in self.test_modules:
            count -= 1
            self.report.log("Deleting module '%s' (tenant: %s)" % (
                test_module.name, test_module.tenant_id))
            self.assert_module_delete(self.admin_client, test_module.id, count)
        self.assert_not_equal(start_count, count, "Nothing was deleted")
        count = len(self.admin_client.modules.list())
        self.assert_equal(self.module_admin_count_prior_to_create, count,
                          "Wrong number of admin modules after deleting all")
        count = len(self.auth_client.modules.list())
        self.assert_equal(self.module_count_prior_to_create, count,
                          "Wrong number of modules after deleting all")

    def assert_module_delete(self, client, module_id, expected_count):
        client.modules.delete(module_id)
        count = len(client.modules.list())
        self.assert_equal(expected_count, count,
                          "Wrong number of modules after delete")



# Copyright 2013 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from oslo_log import log as logging

from trove.common import exception
from trove.common import policy
from trove.common import wsgi
from trove.datastore import models, views
from trove.flavor import views as flavor_views
from trove.volume_type import views as volume_type_view

LOG = logging.getLogger(__name__)


class DatastoreController(wsgi.Controller):

    @classmethod
    def authorize_request(cls, req, rule_name):
        """Datastores are not owned by any particular tenant so we only check
        the current tenant is allowed to perform the action.
        """
        context = req.environ[wsgi.CONTEXT_KEY]
        policy.authorize_on_tenant(context, 'datastore:%s' % rule_name)

    def show(self, req, tenant_id, id):
        self.authorize_request(req, 'show')
        datastore = models.Datastore.load(id)
        datastore_versions = (models.DatastoreVersions.load(datastore.id))
        return wsgi.Result(views.
                           DatastoreView(datastore, datastore_versions,
                                         req).data(), 200)

    def index(self, req, tenant_id):
        self.authorize_request(req, 'index')
        context = req.environ[wsgi.CONTEXT_KEY]
        only_active = True
        if context.is_admin:
            only_active = False
        datastores = models.Datastores.load(only_active)
        datastores_versions = models.DatastoreVersions.load_all(only_active)
        return wsgi.Result(views.
                           DatastoresView(datastores, datastores_versions,
                                          req).data(), 200)

    def version_show(self, req, tenant_id, datastore, id):
        self.authorize_request(req, 'version_show')
        datastore = models.Datastore.load(datastore)
        datastore_version = models.DatastoreVersion.load(datastore, id)
        return wsgi.Result(views.DatastoreVersionView(datastore_version,
                                                      req).data(), 200)

    def version_show_by_uuid(self, req, tenant_id, uuid):
        self.authorize_request(req, 'version_show_by_uuid')
        datastore_version = models.DatastoreVersion.load_by_uuid(uuid)
        return wsgi.Result(views.DatastoreVersionView(datastore_version,
                                                      req).data(), 200)

    def version_index(self, req, tenant_id, datastore):
        self.authorize_request(req, 'version_index')
        context = req.environ[wsgi.CONTEXT_KEY]
        only_active = True
        if context.is_admin:
            only_active = False
        datastore_versions = models.DatastoreVersions.load(datastore,
                                                           only_active)
        return wsgi.Result(views.
                           DatastoreVersionsView(datastore_versions,
                                                 req).data(), 200)

    def list_associated_flavors(self, req, tenant_id, datastore,
                                version_id):
        """
        All nova flavors are returned for a datastore-version unless
        one or more entries are found in datastore_version_metadata,
        in which case only those are returned.
        """
        self.authorize_request(req, 'list_associated_flavors')
        context = req.environ[wsgi.CONTEXT_KEY]
        flavors = (models.DatastoreVersionMetadata.
                   list_datastore_version_flavor_associations(
                       context, version_id))
        return wsgi.Result(flavor_views.FlavorsView(flavors, req).data(), 200)

    def list_associated_volume_types(self, req, tenant_id, datastore,
                                     version_id):
        """
        Return all known volume types if no restrictions have been
        established in datastore_version_metadata, otherwise return
        that restricted set.
        """
        context = req.environ[wsgi.CONTEXT_KEY]
        volume_types = (models.DatastoreVersionMetadata.
                        allowed_datastore_version_volume_types(
                            context, version_id))
        return wsgi.Result(volume_type_view.VolumeTypesView(
            volume_types, req).data(), 200)

    def delete(self, req, tenant_id, id):
        """Remove an existing datastore."""
        self.authorize_request(req, 'delete')

        ds_versions = models.DatastoreVersions.load(id, only_active=False)
        if len(ds_versions.db_info.all()) > 0:
            raise exception.DatastoreVersionsExist(datastore=id)

        LOG.info("Deleting datastore %s", id)

        datastore = models.Datastore.load(id)
        datastore.delete()
        return wsgi.Result(None, 202)



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import re

import docker
from oslo_log import log as logging
from oslo_utils import encodeutils

from trove.common import cfg

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
ANSI_ESCAPE = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')


def stop_container(client, name="database"):
    try:
        container = client.containers.get(name)
    except docker.errors.NotFound:
        LOG.warning("Failed to get container %s", name)
        return

    container.stop(timeout=CONF.state_change_wait_time)


def start_container(client, image, name="database",
                    restart_policy="unless-stopped",
                    volumes={}, ports={}, user="", network_mode="host",
                    environment={}, command=""):
    """Start a docker container.

    :param client: docker client obj.
    :param image: docker image.
    :param name: container name.
    :param restart_policy: restart policy.
    :param volumes: e.g.
           {"/host/trove": {"bind": "/container/trove", "mode": "rw"}}
    :param ports: ports is ignored when network_mode="host". e.g.
           {"3306/tcp": 3306}
    :param user: e.g. "1000.1001"
    :param network_mode: One of bridge, none, host
    :param environment: Environment variables
    :param command:
    :return:
    """
    try:
        container = client.containers.get(name)
        LOG.info(f'Starting existing container {name}')
        container.start()
    except docker.errors.NotFound:
        LOG.info(
            f"Creating docker container, image: {image}, "
            f"volumes: {volumes}, ports: {ports}, user: {user}, "
            f"network_mode: {network_mode}, environment: {environment}, "
            f"command: {command}")
        container = client.containers.run(
            image,
            name=name,
            restart_policy={"Name": restart_policy},
            privileged=False,
            network_mode=network_mode,
            detach=True,
            volumes=volumes,
            ports=ports,
            user=user,
            environment=environment,
            command=command
        )

    return container


def _decode_output(output):
    output = encodeutils.safe_decode(output)
    output = ANSI_ESCAPE.sub('', output.strip())
    return output.split('\n')


def run_container(client, image, name, network_mode="host", volumes={},
                  command="", user=""):
    """Run command in a container and return the string output list.

    :returns output: The log output.
    :returns ret: True if no error occurs, otherwise False.
    """
    try:
        container = client.containers.get(name)
        LOG.debug(f'Removing existing container {name}')
        container.remove(force=True)
    except docker.errors.NotFound:
        pass

    try:
        LOG.info(
            f'Running container {name}, image: {image}, '
            f'network_mode: {network_mode}, volumes: {volumes}, '
            f'command: {command}')
        output = client.containers.run(
            image,
            name=name,
            network_mode=network_mode,
            volumes=volumes,
            remove=False,
            command=command,
            user=user,
        )
    except docker.errors.ContainerError as err:
        output = err.container.logs()
        return _decode_output(output), False

    return _decode_output(output), True


def get_container_status(client, name="database"):
    try:
        container = client.containers.get(name)
        # One of created, restarting, running, removing, paused, exited, or
        # dead
        return container.status
    except docker.errors.NotFound:
        return "not running"
    except Exception:
        return "unknown"


def run_command(client, command, name="database"):
    container = client.containers.get(name)
    # output is Bytes type
    ret, output = container.exec_run(command)
    if ret == 1:
        raise Exception('Running command error: %s' % output)

    return output


def restart_container(client, name="database"):
    container = client.containers.get(name)
    container.restart(timeout=CONF.state_change_wait_time)


def remove_container(client, name="database"):
    try:
        container = client.containers.get(name)
        container.remove(force=True)
    except docker.errors.NotFound:
        pass


def get_container_logs(client, name='database', tail=50):
    container = client.containers.get(name)
    output = container.logs(tail=tail)
    return _decode_output(output)


def prune_images(client):
    """Remove unused images."""
    try:
        client.images.prune(filters={'dangling': False})
    except Exception as e:
        LOG.warning(f"Prune image failed, error: {str(e)}")



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import test

from trove.tests.scenario import groups
from trove.tests.scenario.groups.test_group import TestGroup
from trove.tests.scenario.runners import test_runners


GROUP = "scenario.instance_create_group"


class InstanceCreateRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'instance_create_runners'
    _runner_cls = 'InstanceCreateRunner'


@test(groups=[GROUP, groups.INST_CREATE])
class InstanceCreateGroup(TestGroup):
    """Test Instance Create functionality."""

    def __init__(self):
        super(InstanceCreateGroup, self).__init__(
            InstanceCreateRunnerFactory.instance())

    @test
    def create_empty_instance(self):
        """Create an empty instance."""
        self.test_runner.run_empty_instance_create()


@test(depends_on_classes=[InstanceCreateGroup],
      groups=[GROUP, groups.INST_INIT_CREATE])
class InstanceInitCreateGroup(TestGroup):
    """Test Instance Init Create functionality."""

    def __init__(self):
        super(InstanceInitCreateGroup, self).__init__(
            InstanceCreateRunnerFactory.instance())

    @test
    def create_initial_configuration(self):
        """Create a configuration group for a new initialized instance."""
        self.test_runner.run_initial_configuration_create()

    @test(runs_after=[create_initial_configuration])
    def create_initialized_instance(self):
        """Create an instance with initial properties."""
        self.test_runner.run_initialized_instance_create()


@test(depends_on_classes=[InstanceCreateGroup],
      groups=[GROUP, groups.INST_CREATE])
class InstanceCreateWaitGroup(TestGroup):
    """Test that Instance Create Completes."""

    def __init__(self):
        super(InstanceCreateWaitGroup, self).__init__(
            InstanceCreateRunnerFactory.instance())

    @test
    def wait_for_instance(self):
        """Waiting for main instance to become active."""
        self.test_runner.run_wait_for_instance()


@test(depends_on_classes=[InstanceCreateWaitGroup],
      groups=[GROUP, groups.INST_INIT_CREATE_WAIT])
class InstanceInitCreateWaitGroup(TestGroup):
    """Test that Instance Init Create Completes."""

    def __init__(self):
        super(InstanceInitCreateWaitGroup, self).__init__(
            InstanceCreateRunnerFactory.instance())

    @test
    def wait_for_init_instance(self):
        """Waiting for init instance to become active."""
        self.test_runner.run_wait_for_init_instance()

    @test(depends_on=[wait_for_init_instance])
    def add_initialized_instance_data(self):
        """Add data to the initialized instance."""
        self.test_runner.run_add_initialized_instance_data()

    @test(runs_after=[add_initialized_instance_data])
    def validate_initialized_instance(self):
        """Validate the initialized instance data and properties."""
        self.test_runner.run_validate_initialized_instance()


@test(depends_on_classes=[InstanceInitCreateWaitGroup],
      groups=[GROUP, groups.INST_INIT_DELETE])
class InstanceInitDeleteGroup(TestGroup):
    """Test Initialized Instance Delete functionality."""

    def __init__(self):
        super(InstanceInitDeleteGroup, self).__init__(
            InstanceCreateRunnerFactory.instance())

    @test
    def delete_initialized_instance(self):
        """Delete the initialized instance."""
        self.test_runner.run_initialized_instance_delete()


@test(depends_on_classes=[InstanceInitDeleteGroup],
      groups=[GROUP, groups.INST_INIT_DELETE_WAIT])
class InstanceInitDeleteWaitGroup(TestGroup):
    """Test that Initialized Instance Delete Completes."""

    def __init__(self):
        super(InstanceInitDeleteWaitGroup, self).__init__(
            InstanceCreateRunnerFactory.instance())

    @test
    def wait_for_init_delete(self):
        """Wait for the initialized instance to be gone."""
        self.test_runner.run_wait_for_init_delete()

    @test(runs_after=[wait_for_init_delete])
    def delete_initial_configuration(self):
        """Delete the initial configuration group."""
        self.test_runner.run_initial_configuration_delete()



# Copyright 2013 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import re

from oslo_db.sqlalchemy import engines
import pexpect
from sqlalchemy.exc import OperationalError
try:
    from sqlalchemy.exc import ResourceClosedError
except ImportError:
    ResourceClosedError = Exception

from trove import tests
from trove.tests.config import CONFIG


def create_mysql_connection(host, user, password):
    connection = CONFIG.mysql_connection_method
    if connection['type'] == "direct":
        return SqlAlchemyConnection(host, user, password)
    elif connection['type'] == "tunnel":
        if 'ssh' not in connection:
            raise RuntimeError("If connection type is 'tunnel' then a "
                               "property 'ssh' is expected.")
        return PexpectMySqlConnection(connection['ssh'], host, user, password)
    else:
        raise RuntimeError("Unknown Bad test configuration for "
                           "mysql_connection_method")


class MySqlConnectionFailure(RuntimeError):

    def __init__(self, msg):
        super(MySqlConnectionFailure, self).__init__(msg)


class MySqlPermissionsFailure(RuntimeError):

    def __init__(self, msg):
        super(MySqlPermissionsFailure, self).__init__(msg)


class SqlAlchemyConnection(object):

    def __init__(self, host, user, password):
        self.host = host
        self.user = user
        self.password = password
        try:
            self.engine = self._init_engine(user, password, host)
        except OperationalError as oe:
            if self._exception_is_permissions_issue(str(oe)):
                raise MySqlPermissionsFailure(oe)
            else:
                raise MySqlConnectionFailure(oe)

    @staticmethod
    def _exception_is_permissions_issue(msg):
        """Assert message cited a permissions issue and not something else."""
        pos_error = re.compile(r".*Host '[\w\.]*' is not allowed to connect "
                               "to this MySQL server.*")
        pos_error1 = re.compile(".*Access denied for user "
                                r"'[\w\*\!\@\#\^\&]*'@'[\w\.]*'.*")
        if (pos_error.match(msg) or pos_error1.match(msg)):
            return True

    def __enter__(self):
        try:
            self.conn = self.engine.connect()
        except OperationalError as oe:
            if self._exception_is_permissions_issue(str(oe)):
                raise MySqlPermissionsFailure(oe)
            else:
                raise MySqlConnectionFailure(oe)
        self.trans = self.conn.begin()
        return self

    def execute(self, cmd):
        """Execute some code."""
        cmd = cmd.replace("%", "%%")
        try:
            return self.conn.execute(cmd).fetchall()
        except Exception:
            self.trans.rollback()
            self.trans = None
            try:
                raise
            except ResourceClosedError:
                return []

    def __exit__(self, type, value, traceback):
        if self.trans:
            if type is not None:  # An error occurred
                self.trans.rollback()
            else:
                self.trans.commit()
        self.conn.close()

    @staticmethod
    def _init_engine(user, password, host):
        return engines.create_engine(
            "mysql+pymysql://%s:%s@%s:3306" % (user, password, host))


class PexpectMySqlConnection(object):

    TIME_OUT = 30

    def __init__(self, ssh_args, host, user, password):
        self.host = host
        self.user = user
        self.password = password
        cmd = '%s %s' % (tests.SSH_CMD, ssh_args)
        self.proc = pexpect.spawn(cmd)
        print(cmd)
        self.proc.expect(r":~\$", timeout=self.TIME_OUT)
        cmd2 = "mysql --host '%s' -u '%s' '-p%s'\n" % \
               (self.host, self.user, self.password)
        print(cmd2)
        self.proc.send(cmd2)
        result = self.proc.expect([
            'mysql>',
            'Access denied',
            "Can't connect to MySQL server"],
            timeout=self.TIME_OUT)
        if result == 1:
            raise MySqlPermissionsFailure(self.proc.before)
        elif result == 2:
            raise MySqlConnectionFailure(self.proc.before)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.proc.close()

    def execute(self, cmd):
        self.proc.send(cmd + "\\G\n")
        outcome = self.proc.expect(['Empty set', 'mysql>'],
                                   timeout=self.TIME_OUT)
        if outcome == 0:
            return []
        else:
            # This next line might be invaluable for long test runs.
            print("Interpreting output: %s" % self.proc.before)
            lines = self.proc.before.split("\r\n")
            result = []
            row = None
            for line in lines:
                plural_s = "s" if len(result) != 0 else ""
                end_line = "%d row%s in set" % ((len(result) + 1), plural_s)
                if len(result) == 0:
                    end_line = "1 row in set"
                if (line.startswith("***************************") or
                        line.startswith(end_line)):
                    if row is not None:
                        result.append(row)
                    row = {}
                elif row is not None:
                    colon = line.find(": ")
                    field = line[:colon]
                    value = line[colon + 2:]
                    row[field] = value
            return result



# Copyright 2021 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from unittest import mock

from trove.common import exception
from trove.common import wsgi
from trove.extensions.mgmt.quota import service as quota_service
from trove.tests.unittests import trove_testtools
from trove.tests.unittests.util import util


class TestQuotaController(trove_testtools.TestCase):
    @classmethod
    def setUpClass(cls):
        util.init_db()
        cls.controller = quota_service.QuotaController()
        cls.admin_project_id = cls.random_uuid()
        super(TestQuotaController, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        util.cleanup_db()
        super(TestQuotaController, cls).tearDownClass()

    def test_show_admin_query(self):
        user_project_id = self.random_uuid()
        req_mock = mock.MagicMock(
            environ={
                wsgi.CONTEXT_KEY: mock.MagicMock(
                    project_id=self.admin_project_id,
                    is_admin=True
                )
            }
        )
        result = self.controller.show(req_mock, self.admin_project_id,
                                      user_project_id)

        self.assertEqual(200, result.status)

    def test_show_user_query(self):
        """Show the tenant's own quota."""
        user_project_id = self.random_uuid()
        req_mock = mock.MagicMock(
            environ={
                wsgi.CONTEXT_KEY: mock.MagicMock(
                    is_admin=False
                )
            }
        )
        result = self.controller.show(req_mock, user_project_id,
                                      user_project_id)

        self.assertEqual(200, result.status)

    def test_show_user_query_not_allowed(self):
        """Show other tenant's quota should fail."""
        user_project_id = self.random_uuid()
        req_mock = mock.MagicMock(
            environ={
                wsgi.CONTEXT_KEY: mock.MagicMock(
                    is_admin=False
                )
            }
        )
        self.assertRaises(
            exception.TroveOperationAuthError,
            self.controller.show,
            req_mock, user_project_id,
            self.random_uuid()
        )



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Various conveniences used for migration scripts."""

from oslo_log import log as logging
import sqlalchemy.types

logger = logging.getLogger('trove.db.sqlalchemy.migrate_repo.schema')


class String(sqlalchemy.types.String):
    def __init__(self, length, *args, **kwargs):
        super(String, self).__init__(*args, length=length, **kwargs)


class Text(sqlalchemy.types.Text):
    def __init__(self, length=None, *args, **kwargs):
        super(Text, self).__init__(*args, **kwargs)
        self.with_variant(sqlalchemy.types.Text(length=length), 'mysql')


class Boolean(sqlalchemy.types.Boolean):
    def __init__(self, create_constraint=True, name=None, *args, **kwargs):
        super(Boolean, self).__init__(*args,
                                      create_constraint=create_constraint,
                                      name=name,
                                      **kwargs)


class DateTime(sqlalchemy.types.DateTime):
    def __init__(self, timezone=False, *args, **kwargs):
        super(DateTime, self).__init__(*args,
                                       timezone=timezone,
                                       **kwargs)


class Integer(sqlalchemy.types.Integer):
    def __init__(self, *args, **kwargs):
        super(Integer, self).__init__(*args, **kwargs)


class BigInteger(sqlalchemy.types.BigInteger):
    def __init__(self, *args, **kwargs):
        super(BigInteger, self).__init__(*args, **kwargs)


class Float(sqlalchemy.types.Float):
    def __init__(self, *args, **kwargs):
        super(Float, self).__init__(*args, **kwargs)


def create_tables(tables):
    for table in tables:
        logger.info("creating table %(table)s", {'table': table})
        table.create()


def drop_tables(tables):
    for table in tables:
        logger.info("dropping table %(table)s", {'table': table})
        table.drop()


def Table(name, metadata, *args, **kwargs):
    return sqlalchemy.schema.Table(name, metadata, *args,
                                   mysql_engine='INNODB', **kwargs)



# Copyright 2020 Catalyst Cloud
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy.schema import Column
from sqlalchemy.schema import Index
from sqlalchemy.schema import MetaData
from sqlalchemy.schema import UniqueConstraint

from trove.db.sqlalchemy.migrate_repo.schema import create_tables
from trove.db.sqlalchemy.migrate_repo.schema import DateTime
from trove.db.sqlalchemy.migrate_repo.schema import String
from trove.db.sqlalchemy.migrate_repo.schema import Table

meta = MetaData()

backup_strategy = Table(
    'backup_strategy',
    meta,
    Column('id', String(36), primary_key=True, nullable=False),
    Column('tenant_id', String(36), nullable=False),
    Column('instance_id', String(36), nullable=False, default=''),
    Column('backend', String(255), nullable=False),
    Column('swift_container', String(255), nullable=True),
    Column('created', DateTime()),
    UniqueConstraint(
        'tenant_id', 'instance_id',
        name='UQ_backup_strategy_tenant_id_instance_id'),
    Index("backup_strategy_tenant_id_instance_id", "tenant_id", "instance_id"),
)


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    create_tables([backup_strategy])



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import random
import testtools
from unittest import mock
import uuid

from trove.common import cfg
from trove.common.context import TroveContext
from trove.common.notification import DBaaSAPINotification
from trove.common import policy
from trove.tests import root_logger


def is_bool(val):
    return str(val).lower() in ['true', '1', 't', 'y', 'yes', 'on', 'set']


def patch_notifier(test_case):
    notification_notify = mock.patch.object(
        DBaaSAPINotification, "_notify")
    notification_notify.start()
    test_case.addCleanup(notification_notify.stop)


class TroveTestNotification(DBaaSAPINotification):

    @abc.abstractmethod
    def event_type(self):
        return 'test_notification'

    @abc.abstractmethod
    def required_start_traits(self):
        return []


class TroveTestContext(TroveContext):

    def __init__(self, test_case, **kwargs):
        super(TroveTestContext, self).__init__(**kwargs)
        self.notification = TroveTestNotification(
            self, request_id='req_id', flavor_id='7')
        self.notification.server_type = 'api'
        patch_notifier(test_case)


class TestCase(testtools.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestCase, cls).setUpClass()
        root_logger.DefaultRootLogger(enable_backtrace=False)

    def setUp(self):
        super(TestCase, self).setUp()

        self.addCleanup(cfg.CONF.reset)

        root_logger.DefaultRootHandler.set_info(self.id())

        # Default manager used by all unittsest unless explicitly overridden.
        self.patch_datastore_manager('mysql')

        policy_patcher = mock.patch.object(policy, 'get_enforcer',
                                           return_value=mock.MagicMock())
        self.addCleanup(policy_patcher.stop)
        policy_patcher.start()

    def tearDown(self):
        # yes, this is gross and not thread aware.
        # but the only way to make it thread aware would require that
        # we single thread all testing
        root_logger.DefaultRootHandler.set_info(info=None)
        super(TestCase, self).tearDown()

    def patch_datastore_manager(self, manager_name):
        return self.patch_conf_property('datastore_manager', manager_name)

    def patch_conf_property(self, property_name, value, section=None):
        target = cfg.CONF
        if section:
            target = target.get(section)
        conf_patcher = mock.patch.object(
            target, property_name,
            new_callable=mock.PropertyMock(return_value=value))
        self.addCleanup(conf_patcher.stop)
        return conf_patcher.start()

    @classmethod
    def random_name(cls, name='', prefix=None):
        """Generate a random name that inclues a random number.

        :param str name: The name that you want to include
        :param str prefix: The prefix that you want to include

        :return: a random name. The format is
                 '<prefix>-<name>-<random number>'.
                 (e.g. 'prefixfoo-namebar-154876201')
        :rtype: string
        """
        randbits = str(random.randint(1, 0x7fffffff))
        rand_name = randbits
        if name:
            rand_name = name + '-' + rand_name
        if prefix:
            rand_name = prefix + '-' + rand_name
        return rand_name

    @classmethod
    def random_uuid(cls):
        return str(uuid.uuid4())

    def assertDictContains(self, parent, child):
        """Checks whether child dict is a subset of parent.

        assertDictContainsSubset() in standard Python 2.7 has been deprecated
        since Python 3.2
        """
        self.assertEqual(parent, dict(parent, **child))



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import signal
import subprocess

from oslo_config import cfg
from oslo_log import log as logging

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class BaseRunner(object):
    """Base class for Backup Strategy implementations."""

    # Subclass should provide the commands.
    cmd = ''
    restore_cmd = ''
    prepare_cmd = ''

    encrypt_key = CONF.backup_encryption_key

    def __init__(self, *args, **kwargs):
        self.process = None
        self.pid = None
        self.base_filename = kwargs.get('filename')
        self.storage = kwargs.pop('storage', None)
        self.location = kwargs.pop('location', '')
        self.checksum = kwargs.pop('checksum', '')

        if 'restore_location' not in kwargs:
            kwargs['restore_location'] = self.datadir
        self.restore_location = kwargs['restore_location']
        self.restore_content_length = 0

        self.command = self.cmd % kwargs

        if self.location.endswith('.enc') and not self.encrypt_key:
            raise Exception("Encryption key not provided with an encrypted "
                            "backup.")

        self.restore_command = ''
        # Only decrypt if the object name ends with .enc
        if self.location.endswith('.enc'):
            self.restore_command = self.decrypt_cmd
        self.restore_command = (self.restore_command +
                                self.unzip_cmd +
                                (self.restore_cmd % kwargs))
        self.prepare_command = self.prepare_cmd % kwargs

    @property
    def filename(self):
        """Subclasses may overwrite this to declare a format (.tar)."""
        return self.base_filename

    @property
    def manifest(self):
        """Target file name."""
        return "%s%s%s" % (self.filename,
                           self.zip_manifest,
                           self.encrypt_manifest)

    @property
    def zip_cmd(self):
        return ' | gzip'

    @property
    def unzip_cmd(self):
        return 'gzip -d -c | '

    @property
    def zip_manifest(self):
        return '.gz'

    @property
    def encrypt_cmd(self):
        """Encryption command.

        Since Victoria, trove no longer encrypts the backup data for the end
        user. This could be improved by giving users the capability to specify
        password when creating the backups.
        """
        return ""

    @property
    def decrypt_cmd(self):
        """Decryption command.

        Since Victoria, trove no longer encrypts the backup data for the end
        user. This command is only for backward compatibility.
        """
        if self.encrypt_key:
            return ('openssl enc -d -aes-256-cbc -md sha512 -pbkdf2 -iter '
                    '10000 -salt -pass pass:%s | '
                    % self.encrypt_key)
        else:
            return ''

    @property
    def encrypt_manifest(self):
        return '.enc' if self.encrypt_key else ''

    def _run(self):
        LOG.info("Running backup cmd: %s", self.command)
        self.process = subprocess.Popen(self.command, shell=True,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        preexec_fn=os.setsid)
        self.pid = self.process.pid

    def __enter__(self):
        """Start up the process."""
        self.pre_backup()
        self._run()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Clean up everything."""
        if getattr(self, 'process', None):
            try:
                # Send a sigterm to the session leader, so that all
                # child processes are killed and cleaned up on terminate
                os.killpg(self.process.pid, signal.SIGTERM)
                self.process.terminate()
            except OSError:
                pass

            if exc_type is not None:
                return False

            try:
                err = self.process.stderr.read()
                if err:
                    raise Exception(err)
            except OSError:
                pass

            if not self.check_process():
                raise Exception()

        self.post_backup()

        return True

    def read(self, chunk_size):
        return self.process.stdout.read(chunk_size)

    def get_metadata(self):
        """Hook for subclasses to get metadata from the backup."""
        return {}

    def check_process(self):
        """Hook for subclasses to check process for errors."""
        return True

    def check_restore_process(self):
        """Hook for subclasses to check the restore process for errors."""
        return True

    def pre_backup(self):
        """Hook for subclasses to run commands before backup."""
        pass

    def post_backup(self):
        """Hook for subclasses to run commands after backup."""
        pass

    def pre_restore(self):
        """Hook that is called before the restore command."""
        pass

    def post_restore(self):
        """Hook that is called after the restore command."""
        pass

    def unpack(self, location, checksum, command):
        stream = self.storage.load(location, checksum)

        LOG.info('Running restore from stream, command: %s', command)
        self.process = subprocess.Popen(command, shell=True,
                                        stdin=subprocess.PIPE,
                                        stderr=subprocess.PIPE)
        content_length = 0
        for chunk in stream:
            self.process.stdin.write(chunk)
            content_length += len(chunk)
        self.process.stdin.close()

        try:
            err = self.process.stderr.read()
            if err:
                raise Exception(err)
        except OSError:
            pass

        if not self.check_restore_process():
            raise Exception()

        return content_length

    def run_restore(self):
        return self.unpack(self.location, self.checksum, self.restore_command)

    def restore(self):
        """Restore backup to data directory.

        :returns Restored data size.
        """
        self.pre_restore()
        content_length = self.run_restore()
        self.post_restore()
        return content_length



# Copyright 2010-2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Model classes that extend the instances functionality for MySQL instances.
"""

from trove.common import cfg
from trove.common.clients import create_guest_client
from trove.common.db.mysql import models as guest_models
from trove.common import exception
from trove.common.notification import StartNotification
from trove.common import utils
from trove.extensions.common.models import load_and_verify
from trove.extensions.common.models import RootHistory

CONF = cfg.CONF


def persisted_models():
    return {'root_enabled_history': RootHistory}


class User(object):

    _data_fields = ['name', 'host', 'password', 'databases']

    def __init__(self, name, host, password, databases):
        self.name = name
        self.host = host
        self.password = password
        self.databases = databases

    @classmethod
    def load(cls, context, instance_id, username, hostname, root_user=False):
        load_and_verify(context, instance_id)
        validate = guest_models.MySQLUser(name=username, host=hostname)
        if root_user:
            validate.make_root()
        validate.check_reserved()
        client = create_guest_client(context, instance_id)
        found_user = client.get_user(username=username, hostname=hostname)
        if not found_user:
            return None
        database_names = [{'name': db['_name']}
                          for db in found_user['_databases']]
        return cls(found_user['_name'],
                   found_user['_host'],
                   found_user['_password'],
                   database_names)

    @classmethod
    def create(cls, context, instance_id, users):
        # Load InstanceServiceStatus to verify if it's running
        load_and_verify(context, instance_id)
        client = create_guest_client(context, instance_id)
        for user in users:
            user_name = user['_name']
            host_name = user['_host']
            userhost = "%s@%s" % (user_name, host_name)
            existing_users, _nadda = Users.load_with_client(
                client,
                limit=1,
                marker=userhost,
                include_marker=True)
            if (len(existing_users) > 0 and
                    str(existing_users[0].name) == str(user_name) and
                    str(existing_users[0].host) == str(host_name)):
                raise exception.UserAlreadyExists(name=user_name,
                                                  host=host_name)
        return client.create_user(users)

    @classmethod
    def delete(cls, context, instance_id, user):
        load_and_verify(context, instance_id)

        with StartNotification(context, instance_id=instance_id,
                               username=user):
            create_guest_client(context, instance_id).delete_user(user)

    @classmethod
    def access(cls, context, instance_id, username, hostname):
        load_and_verify(context, instance_id)
        client = create_guest_client(context, instance_id)
        databases = client.list_access(username, hostname)
        dbs = []
        for db in databases:
            dbs.append(Schema(name=db['_name'],
                              collate=db['_collate'],
                              character_set=db['_character_set']))
        return UserAccess(dbs)

    @classmethod
    def grant(cls, context, instance_id, username, hostname, databases):
        load_and_verify(context, instance_id)
        client = create_guest_client(context, instance_id)
        client.grant_access(username, hostname, databases)

    @classmethod
    def revoke(cls, context, instance_id, username, hostname, database):
        load_and_verify(context, instance_id)
        client = create_guest_client(context, instance_id)
        client.revoke_access(username, hostname, database)

    @classmethod
    def change_password(cls, context, instance_id, users):
        load_and_verify(context, instance_id)
        client = create_guest_client(context, instance_id)
        change_users = []
        for user in users:
            change_user = {'name': user.name,
                           'host': user.host,
                           'password': user.password,
                           }
            change_users.append(change_user)
        client.change_passwords(change_users)

    @classmethod
    def update_attributes(cls, context, instance_id, username, hostname,
                          user_attrs):
        load_and_verify(context, instance_id)
        client = create_guest_client(context, instance_id)

        user_changed = user_attrs.get('name')
        host_changed = user_attrs.get('host')

        user = user_changed or username
        host = host_changed or hostname

        validate = guest_models.MySQLUser(name=user, host=host)
        validate.check_reserved()

        userhost = "%s@%s" % (user, host)
        if user_changed or host_changed:
            existing_users, _nadda = Users.load_with_client(
                client,
                limit=1,
                marker=userhost,
                include_marker=True)
            if (len(existing_users) > 0 and
                    existing_users[0].name == user and
                    existing_users[0].host == host):
                raise exception.UserAlreadyExists(name=user,
                                                  host=host)
        client.update_attributes(username, hostname, user_attrs)


class UserAccess(object):
    _data_fields = ['databases']

    def __init__(self, databases):
        self.databases = databases


def load_via_context(cls, context, instance_id):
    """Creates guest and fetches pagination arguments from the context."""
    load_and_verify(context, instance_id)
    limit = utils.pagination_limit(context.limit, cls.DEFAULT_LIMIT)
    client = create_guest_client(context, instance_id)
    # The REST API standard dictates that we *NEVER* include the marker.
    return cls.load_with_client(client=client, limit=limit,
                                marker=context.marker, include_marker=False)


class Users(object):

    DEFAULT_LIMIT = CONF.users_page_size

    @classmethod
    def load(cls, context, instance_id):
        return load_via_context(cls, context, instance_id)

    @classmethod
    def load_with_client(cls, client, limit, marker, include_marker):
        user_list, next_marker = client.list_users(
            limit=limit,
            marker=marker,
            include_marker=include_marker)
        model_users = []
        for user in user_list:
            mysql_user = guest_models.MySQLUser.deserialize(user,
                                                            verify=False)
            if mysql_user.name in cfg.get_ignored_users():
                continue
            # TODO(hub-cap): databases are not being returned in the
            # reference agent
            dbs = []
            for db in mysql_user.databases:
                dbs.append({'name': db['_name']})
            model_users.append(User(mysql_user.name,
                                    mysql_user.host,
                                    mysql_user.password,
                                    dbs))
        return model_users, next_marker


class Schema(object):

    _data_fields = ['name', 'collate', 'character_set']

    def __init__(self, name, collate, character_set):
        self.name = name
        self.collate = collate
        self.character_set = character_set

    @classmethod
    def create(cls, context, instance_id, schemas):
        load_and_verify(context, instance_id)
        client = create_guest_client(context, instance_id)
        for schema in schemas:
            schema_name = schema['_name']
            existing_schema, _nadda = Schemas.load_with_client(
                client,
                limit=1,
                marker=schema_name,
                include_marker=True)
            if (len(existing_schema) > 0 and
                    str(existing_schema[0].name) == str(schema_name)):
                raise exception.DatabaseAlreadyExists(name=schema_name)
        return client.create_database(schemas)

    @classmethod
    def delete(cls, context, instance_id, schema):
        load_and_verify(context, instance_id)
        create_guest_client(context, instance_id).delete_database(schema)


class Schemas(object):

    DEFAULT_LIMIT = CONF.databases_page_size

    @classmethod
    def load(cls, context, instance_id):
        return load_via_context(cls, context, instance_id)

    @classmethod
    def load_with_client(cls, client, limit, marker, include_marker):
        schemas, next_marker = client.list_databases(
            limit=limit,
            marker=marker,
            include_marker=include_marker)
        model_schemas = []
        for schema in schemas:
            mysql_schema = guest_models.MySQLSchema.deserialize(schema,
                                                                verify=False)
            if mysql_schema.name in cfg.get_ignored_dbs():
                continue
            model_schemas.append(Schema(mysql_schema.name,
                                        mysql_schema.collate,
                                        mysql_schema.character_set))
        return model_schemas, next_marker

    @classmethod
    def find(cls, context, instance_id, schema_id):
        load_and_verify(context, instance_id)
        client = create_guest_client(context, instance_id)
        model_schemas, _ = cls.load_with_client(client, 1, schema_id, True)
        if model_schemas and model_schemas[0].name == schema_id:
            return model_schemas[0]

        return None



# Copyright 2016 Tesora, Inc
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Model classes that form the core of volume-support functionality"""

from cinderclient import exceptions as cinder_exception
from trove.common import clients
from trove.common import exception as trove_exception
from trove.common import models


class VolumeType(object):

    _data_fields = ['id', 'name', 'is_public', 'description']

    def __init__(self, volume_type=None):
        """Initialize a cinder client volume_type object"""
        self.volume_type = volume_type

    @classmethod
    def load(cls, volume_type_id, context=None, client=None):
        if not(client or context):
            raise trove_exception.InvalidModelError(
                "client or context must be provided to load a volume_type")
        if not client:
            client = clients.create_cinder_client(context)
        try:
            volume_type = client.volume_types.get(volume_type_id)
        except cinder_exception.NotFound:
            raise trove_exception.NotFound(uuid=volume_type_id)
        except cinder_exception.ClientException as ce:
            raise trove_exception.TroveError(str(ce))
        return cls(volume_type)

    @property
    def id(self):
        return self.volume_type.id

    @property
    def name(self):
        return self.volume_type.name

    @property
    def is_public(self):
        return self.volume_type.is_public

    @property
    def description(self):
        return self.volume_type.description


class VolumeTypes(models.CinderRemoteModelBase):

    def __init__(self, context):
        volume_types = clients.create_cinder_client(
            context).volume_types.list()
        self.volume_types = [VolumeType(volume_type=item)
                             for item in volume_types]

    def __iter__(self):
        for item in self.volume_types:
            yield item



# Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy.schema import Column
from sqlalchemy.schema import MetaData

from trove.db.sqlalchemy.migrate_repo.schema import String
from trove.db.sqlalchemy.migrate_repo.schema import Table


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine
    instances = Table('instances', meta, autoload=True)
    service_type = Column('service_type', String(36))
    instances.create_column(service_type)
    instances.update().values({'service_type': 'mysql'}).execute()



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import ast
import csv
import configparser
import io
import re
import sys

import xmltodict
import yaml

from oslo_serialization import base64
from oslo_serialization import jsonutils
from oslo_utils import strutils

from trove.common import utils as trove_utils


class StringConverter(object):
    """A passthrough string-to-object converter.
    """

    def __init__(self, object_mappings):
        """
        :param object_mappings:  string-to-object mappings
        :type object_mappings:   dict
        """
        self._object_mappings = object_mappings

    def to_strings(self, items):
        """Recursively convert collection items to strings.

        :returns:        Copy of the input collection with all items converted.
        """
        if trove_utils.is_collection(items):
            return map(self.to_strings, items)

        return self._to_string(items)

    def to_objects(self, items):
        """Recursively convert collection string to objects.

        :returns:        Copy of the input collection with all items converted.
        """
        if trove_utils.is_collection(items):
            return map(self.to_objects, items)

        return self._to_object(items)

    def _to_string(self, value):
        for k, v in self._object_mappings.items():
            if v is value:
                return k

        return str(value)

    def _to_object(self, value):
        # Return known mappings and quoted strings right away.
        if value in self._object_mappings:
            return self._object_mappings[value]
        elif (isinstance(value, str) and
              re.match("^'(.*)'|\"(.*)\"$", value)):
            return value

        try:
            return ast.literal_eval(value)
        except Exception:
            return value


class StreamCodec(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def serialize(self, data):
        """Serialize a Python object into a stream.
        """

    @abc.abstractmethod
    def deserialize(self, stream):
        """Deserialize stream data into a Python structure.
        """


class IdentityCodec(StreamCodec):
    """
    A basic passthrough codec.
    Does not modify the data in any way.
    """

    def serialize(self, data):
        return data

    def deserialize(self, stream):
        return stream


class YamlCodec(StreamCodec):
    """
    Read/write data from/into a YAML config file.

    a: 1
    b: {c: 3, d: 4}
    ...

    The above file content (flow-style) would be represented as:
    {'a': 1,
     'b': {'c': 3, 'd': 4,}
     ...
    }
    """

    def __init__(self, default_flow_style=False):
        """
        :param default_flow_style:  Use flow-style (inline) formatting of
                                    nested collections.
        :type default_flow_style:   boolean
        """
        self._default_flow_style = default_flow_style

    def serialize(self, dict_data):
        return yaml.dump(dict_data, Dumper=self.dumper,
                         default_flow_style=self._default_flow_style)

    def deserialize(self, stream):
        return yaml.load(stream, Loader=self.loader)

    @property
    def loader(self):
        return yaml.loader.Loader

    @property
    def dumper(self):
        return yaml.dumper.Dumper


class SafeYamlCodec(YamlCodec):
    """
    Same as YamlCodec except that it uses safe Loader and Dumper which
    encode Unicode strings and produce only basic YAML tags.
    """

    def __init__(self, default_flow_style=False):
        super(SafeYamlCodec, self).__init__(
            default_flow_style=default_flow_style)

    @property
    def loader(self):
        return yaml.loader.SafeLoader

    @property
    def dumper(self):
        return yaml.dumper.SafeDumper


class IniCodec(StreamCodec):
    """
    Read/write data from/into an ini-style config file.

    [section_1]
    key = value
    key = value
    ...

    [section_2]
    key = value
    key = value
    ...

    The above file content would be represented as:
    {'section_1': {'key': value, 'key': value, ...},
     'section_2': {'key': value, 'key': value, ...}
     ...
    }
    """

    def __init__(self, default_value=None, comment_markers=('#', ';')):
        """
        :param default_value:  Default value for keys with no value.
                               If set, all keys are written as 'key = value'.
                               The key is written without trailing '=' if None.
        :type default_value:   object
        """
        self._default_value = default_value
        self._comment_markers = comment_markers

    def serialize(self, dict_data):
        parser = self._init_config_parser(dict_data)
        output = io.StringIO()
        parser.write(output)

        return output.getvalue()

    def deserialize(self, stream):
        parser = self._init_config_parser()
        parser.read_file(self._pre_parse(stream))

        return {s: {k:
                    StringConverter({None: self._default_value}).to_objects(v)
                    for k, v in parser.items(s, raw=True)}
                for s in parser.sections()}

    def _pre_parse(self, stream):
        buf = io.StringIO()
        for line in io.StringIO(stream):
            # Ignore commented lines.
            if not line.startswith(self._comment_markers):
                # Strip leading and trailing whitespaces from each line.
                buf.write(line.strip() + '\n')

        # Rewind the output buffer.
        buf.flush()
        buf.seek(0)

        return buf

    def _init_config_parser(self, sections=None):
        # SafeConfigParser was deprecated in Python 3.2
        if sys.version_info >= (3, 2):
            parser = configparser.ConfigParser(allow_no_value=True)
        else:
            parser = configparser.SafeConfigParser(allow_no_value=True)
        if sections:
            for section in sections:
                parser.add_section(section)
                for key, value in sections[section].items():
                    str_val = StringConverter(
                        {self._default_value: None}).to_strings(value)
                    parser.set(section, key,
                               str(str_val) if str_val is not None
                               else str_val)

        return parser


class PropertiesCodec(StreamCodec):
    """
    Read/write data from/into a property-style config file.

    key1 k1arg1 k1arg2 ... k1argN
    key2 k2arg1 k2arg2 ... k2argN
    key3 k3arg1 k3arg2 ...
    key3 k3arg3 k3arg4 ...
    ...

    The above file content would be represented as:
    {'key1': [k1arg1, k1arg2 ... k1argN],
     'key2': [k2arg1, k2arg2 ... k2argN]
     'key3': [[k3arg1, k3arg2, ...], [k3arg3, k3arg4, ...]]
     ...
    }
    """

    QUOTING_MODE = csv.QUOTE_MINIMAL
    STRICT_MODE = False
    SKIP_INIT_SPACE = True

    def __init__(self, delimiter=' ', comment_markers=('#'),
                 unpack_singletons=True, string_mappings=None):
        """
        :param delimiter:         A one-character used to separate fields.
        :type delimiter:          string

        :param empty_value:       Value to represent None in the output.
        :type empty_value:        object

        :param comment_markers:   List of comment markers.
        :type comment_markers:    list

        :param unpack_singletons: Whether to unpack singleton collections
                                  (collections with only a single item).
        :type unpack_singletons:  boolean

        :param string_mappings:   User-defined string representations of
                                  Python objects.
        :type string_mappings:    dict
        """
        self._delimiter = delimiter
        self._comment_markers = comment_markers
        self._string_converter = StringConverter(string_mappings or {})
        self._unpack_singletons = unpack_singletons

    def serialize(self, dict_data):
        output = io.StringIO()
        writer = csv.writer(output, delimiter=self._delimiter,
                            quoting=self.QUOTING_MODE,
                            strict=self.STRICT_MODE,
                            skipinitialspace=self.SKIP_INIT_SPACE)

        for key, value in dict_data.items():
            writer.writerows(self._to_rows(key, value))

        return output.getvalue()

    def deserialize(self, stream):
        reader = csv.reader(io.StringIO(stream),
                            delimiter=self._delimiter,
                            quoting=self.QUOTING_MODE,
                            strict=self.STRICT_MODE,
                            skipinitialspace=self.SKIP_INIT_SPACE)

        return self._to_dict(reader)

    def _to_dict(self, reader):
        data_dict = {}
        for row in reader:
            if row:
                key = row[0].strip()
                # Ignore comment lines.
                if not key.strip().startswith(self._comment_markers):
                    # NOTE(zhaochao): a list object is expected for
                    # trove_utils.unpack_singleton, however in python3
                    # map objects won't be treated as lists, so we
                    # convert the result of StringConverter.to_objects
                    # to a list explicitly.
                    items = list(self._string_converter.to_objects(
                        [v if v else None for v in
                         map(self._strip_comments, row[1:])]))
                    current = data_dict.get(key)
                    if current is not None:
                        current.append(trove_utils.unpack_singleton(items)
                                       if self._unpack_singletons else items)
                    else:
                        data_dict.update({key: [items]})

        if self._unpack_singletons:
            # Unpack singleton values.
            # NOTE(zhaochao): In Python 3.x, dict.items() returns a view
            # object, which will reflect the changes of the dict itself:
            # https://docs.python.org/3/library/stdtypes.html#dict-views
            # This means as we're changing the dict, dict.items() cannot
            # guarantee we're safely iterating all entries in the dict.
            # Manually converting the result of dict.items() to a list will
            # fix.
            for k, v in list(data_dict.items()):
                data_dict.update({k: trove_utils.unpack_singleton(v)})

        return data_dict

    def _strip_comments(self, value):
        # Strip in-line comments.
        for marker in self._comment_markers:
            value = value.split(marker)[0]
        return value.strip()

    def _to_rows(self, header, items):
        rows = []
        if trove_utils.is_collection(items):
            if any(trove_utils.is_collection(item) for item in items):
                # This is multi-row property.
                for item in items:
                    rows.extend(self._to_rows(header, item))
            else:
                # This is a single-row property with multiple arguments.
                rows.append(self._to_list(
                    header, self._string_converter.to_strings(items)))
        else:
            # This is a single-row property with only one argument.
            # Note(zhaochao): csv.writerows expects a list object before
            # python 3.5, but map objects won't be treated as lists in
            # python 3, so we explicitly convert the result of
            # StringConverter.to_strings to a list here to support py34
            # unittests.
            rows.append(
                list(self._string_converter.to_strings(
                    self._to_list(header, items))))

        return rows

    def _to_list(self, *items):
        container = []
        for item in items:
            if trove_utils.is_collection(item):
                # This item is a nested collection - unpack it.
                container.extend(self._to_list(*item))
            else:
                # This item is not a collection - append it to the list.
                container.append(item)

        return container


class KeyValueCodec(StreamCodec):
    """
    Read/write data from/into a simple key=value file.

    key1=value1
    key2=value2
    key3=value3
    ...

    The above file content would be represented as:
    {'key1': 'value1',
     'key2': 'value2',
     'key3': 'value3',
     ...
    }
    """

    BOOL_PYTHON = 0  # True, False
    BOOL_LOWER = 1   # true, false
    BOOL_UPPER = 2   # TRUE, FALSE

    def __init__(self, delimiter='=',
                 comment_marker='#',
                 line_terminator='\r\n',
                 value_quoting=False,
                 value_quote_char="'",
                 bool_case=BOOL_PYTHON,
                 big_ints=False,
                 hidden_marker=None):
        """
        :param delimiter:        string placed between key and value
        :param comment_marker:   string indicating comment line in file
        :param line_terminator:  string placed between lines
        :param value_quoting:    whether or not to quote string values
        :param value_quote_char: character used to quote string values
        :param bool_case:        BOOL_* setting case of bool values
        :param big_ints:         treat K/M/G at the end of ints as an int
        :param hidden_marker:    pattern prefixing hidden param
        """
        self._delimeter = delimiter
        self._comment_marker = comment_marker
        self._line_terminator = line_terminator
        self._value_quoting = value_quoting
        self._value_quote_char = value_quote_char
        self._bool_case = bool_case
        self._big_ints = big_ints
        self._hidden_marker = hidden_marker

    def serialize(self, dict_data):
        lines = []
        for k, v in dict_data.items():
            lines.append(k + self._delimeter + self.serialize_value(v))
        return self._line_terminator.join(lines)

    def deserialize(self, stream):
        # Note(zhaochao): In Python 3, when files are opened in text mode,
        # newlines will be translated to '\n' by default, so we just split
        # the stream by '\n'.
        if sys.version_info[0] >= 3:
            lines = stream.split('\n')
        else:
            lines = stream.split(self._line_terminator)

        result = {}
        for line in lines:
            line = line.lstrip().rstrip()
            if line == '' or line.startswith(self._comment_marker):
                continue
            k, v = re.split(re.escape(self._delimeter), line, 1)
            if self._value_quoting and v.startswith(self._value_quote_char):
                # remove trailing comments
                v = re.sub(r'%s *%s.*$' % ("'", '#'), '', v)
                v = v.lstrip(
                    self._value_quote_char).rstrip(
                    self._value_quote_char)
            elif v.lower() in ['true', 'false']:
                v = strutils.bool_from_string(v.lower())
            else:
                # remove trailing comments
                v = re.sub('%s.*$' % self._comment_marker, '', v)
            if self._hidden_marker and v.startswith(self._hidden_marker):
                continue
            result[k.strip()] = v
        return result

    def serialize_value(self, value):
        if isinstance(value, bool):
            if self._bool_case == self.BOOL_PYTHON:
                value = str(value)
            elif self._bool_case == self.BOOL_LOWER:
                value = str(value).lower()
            elif self._bool_case == self.BOOL_UPPER:
                value = str(value).upper()
        if self.should_quote_value(value):
            value = self._value_quote_char + value + self._value_quote_char
        return str(value)

    def should_quote_value(self, value):
        if not self._value_quoting:
            return False
        if isinstance(value, bool) or isinstance(value, int):
            return False
        if value.lower() in ['true', 'false']:
            return False
        try:
            int(value)
            return False
        except ValueError:
            pass
        if self._big_ints and re.match(r'\d+[kKmMgGtTpP]', value):
            return False
        return True


class JsonCodec(StreamCodec):

    def serialize(self, dict_data):
        return jsonutils.dumps(dict_data)

    def deserialize(self, stream):
        if type(stream) == str:
            return jsonutils.load(io.StringIO(stream))
        if type(stream) == bytes:
            return jsonutils.load(io.BytesIO(stream))


class Base64Codec(StreamCodec):
    """Serialize (encode) and deserialize (decode) using the base64 codec.
    To read binary data from a file and b64encode it, used the decode=False
    flag on operating_system's read calls.  Use encode=False to decode
    binary data before writing to a file as well.
    """

    # NOTE(zhaochao): migrate to oslo_serialization.base64 to serialize(return
    # a text object) and deserialize(return a bytes object) data.

    def serialize(self, data):
        return base64.encode_as_text(data)

    def deserialize(self, stream):
        return base64.decode_as_bytes(stream)


class XmlCodec(StreamCodec):

    def __init__(self, encoding='utf-8'):
        self._encoding = encoding

    def serialize(self, dict_data):
        return xmltodict.unparse(
            dict_data, output=None, encoding=self._encoding, pretty=True)

    def deserialize(self, stream):
        return xmltodict.parse(stream, encoding=self._encoding)



# Copyright 2017 Eayun, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging
from trove.common import cfg
from trove.common import exception
from trove.common.i18n import _
from trove.common import wsgi
from trove.extensions.common.service import DefaultRootController
from trove.extensions.redis.models import RedisRoot
from trove.extensions.redis.views import RedisRootCreatedView
from trove.instance.models import DBInstance

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
MANAGER = CONF.datastore_manager if CONF.datastore_manager else 'redis'


class RedisRootController(DefaultRootController):
    def root_create(self, req, body, tenant_id, instance_id, is_cluster):
        """Enable authentication for a redis instance and its replicas if any
        """
        self._validate_can_perform_action(tenant_id, instance_id, is_cluster,
                                          "enable_root")
        password = DefaultRootController._get_password_from_body(body)
        slave_instances = self._get_slaves(tenant_id, instance_id)
        return self._instance_root_create(req, instance_id, password,
                                          slave_instances)

    def root_delete(self, req, tenant_id, instance_id, is_cluster):
        """Disable authentication for a redis instance and its replicas if any
        """
        self._validate_can_perform_action(tenant_id, instance_id, is_cluster,
                                          "disable_root")
        slave_instances = self._get_slaves(tenant_id, instance_id)
        return self._instance_root_delete(req, instance_id, slave_instances)

    def _instance_root_create(self, req, instance_id, password,
                              slave_instances=None):
        LOG.info("Enabling authentication for instance '%s'.",
                 instance_id)
        LOG.info("req : '%s'\n\n", req)
        context = req.environ[wsgi.CONTEXT_KEY]

        original_auth_password = self._get_original_auth_password(
            context, instance_id)

        # Do root-enable and roll back once if operation fails.
        try:
            root = RedisRoot.create(context, instance_id, password)
            if not password:
                password = root.password
        except exception.TroveError:
            self._rollback_once(req, instance_id, original_auth_password)
            raise exception.TroveError(
                _("Failed to do root-enable for instance "
                  "'%(instance_id)s'.") % {'instance_id': instance_id}
            )

        failed_slaves = []
        for slave_id in slave_instances:
            try:
                LOG.info("Enabling authentication for slave instance "
                         "'%s'.", slave_id)
                RedisRoot.create(context, slave_id, password)
            except exception.TroveError:
                failed_slaves.append(slave_id)

        return wsgi.Result(
            RedisRootCreatedView(root, failed_slaves).data(), 200)

    def _instance_root_delete(self, req, instance_id, slave_instances=None):
        LOG.info("Disabling authentication for instance '%s'.",
                 instance_id)
        LOG.info("req : '%s'\n\n", req)
        context = req.environ[wsgi.CONTEXT_KEY]

        is_root_enabled = RedisRoot.load(context, instance_id)
        if not is_root_enabled:
            raise exception.RootHistoryNotFound()

        original_auth_password = self._get_original_auth_password(
            context, instance_id)

        # Do root-disable and roll back once if operation fails.
        try:
            RedisRoot.delete(context, instance_id)
        except exception.TroveError:
            self._rollback_once(req, instance_id, original_auth_password)
            raise exception.TroveError(
                _("Failed to do root-disable for instance "
                  "'%(instance_id)s'.") % {'instance_id': instance_id}
            )

        failed_slaves = []
        for slave_id in slave_instances:
            try:
                LOG.info("Disabling authentication for slave instance "
                         "'%s'.", slave_id)
                RedisRoot.delete(context, slave_id)
            except exception.TroveError:
                failed_slaves.append(slave_id)

        if len(failed_slaves) > 0:
            result = {
                'failed_slaves': failed_slaves
            }
            return wsgi.Result(result, 200)

        return wsgi.Result(None, 204)

    @staticmethod
    def _rollback_once(req, instance_id, original_auth_password):
        LOG.info("Rolling back enable/disable authentication "
                 "for instance '%s'.", instance_id)
        context = req.environ[wsgi.CONTEXT_KEY]
        try:
            if not original_auth_password:
                # Instance never did root-enable before.
                RedisRoot.delete(context, instance_id)
            else:
                # Instance has done root-enable successfully before.
                # So roll back with original password.
                RedisRoot.create(context, instance_id,
                                 original_auth_password)
        except exception.TroveError:
            LOG.exception("Rolling back failed for instance '%s'",
                          instance_id)

    @staticmethod
    def _is_slave(tenant_id, instance_id):
        args = {'id': instance_id, 'tenant_id': tenant_id}
        instance_info = DBInstance.find_by(**args)
        return instance_info.slave_of_id

    @staticmethod
    def _get_slaves(tenant_id, instance_or_cluster_id, deleted=False):
        LOG.info("Getting non-deleted slaves of instance '%s', "
                 "if any.", instance_or_cluster_id)
        args = {'slave_of_id': instance_or_cluster_id, 'tenant_id': tenant_id,
                'deleted': deleted}
        db_infos = DBInstance.find_all(**args)
        slaves = []
        for db_info in db_infos:
            slaves.append(db_info.id)
        return slaves

    @staticmethod
    def _get_original_auth_password(context, instance_id):
        # Check if instance did root-enable before and get original password.
        password = None
        if RedisRoot.load(context, instance_id):
            try:
                password = RedisRoot.get_auth_password(context, instance_id)
            except exception.TroveError:
                raise exception.TroveError(
                    _("Failed to get original auth password of instance "
                      "'%(instance_id)s'.") % {'instance_id': instance_id}
                )
        return password

    def _validate_can_perform_action(self, tenant_id, instance_id, is_cluster,
                                     operation):
        if is_cluster:
            raise exception.ClusterOperationNotSupported(
                operation=operation)

        is_slave = self._is_slave(tenant_id, instance_id)
        if is_slave:
            raise exception.SlaveOperationNotSupported(
                operation=operation)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import jsonschema
from unittest.mock import Mock, MagicMock, patch

from trove.extensions.mgmt.upgrade.models import UpgradeMessageSender
from trove.extensions.mgmt.upgrade.service import UpgradeController
from trove.tests.unittests import trove_testtools


class TestUpgradeController(trove_testtools.TestCase):

    def setUp(self):
        super(TestUpgradeController, self).setUp()
        self.controller = UpgradeController()

        self.body = {
            "upgrade": {
                "instance_id": "27e25b73-88a1-4526-b2b9-919a28b8b33f",
                "instance_version": "v1.0.1",
                "location": "http://swift/trove-guestagent-v1.0.1.tar.gz"}
        }

    def tearDown(self):
        super(TestUpgradeController, self).tearDown()
        self.body = {}

    def _get_validator(self, body):
        """
        Helper method to return a validator
        """
        schema = self.controller.get_schema('create', body)
        return jsonschema.Draft4Validator(schema)

    def test_validate_create(self):
        """
        Test for valid payload in body
        """
        validator = self._get_validator(self.body)
        self.assertTrue(validator.is_valid(self.body))

    def test_validate_create_additional_params(self):
        """
        Test for valid payload with additional params
        """
        self.body["upgrade"]["description"] = "upgrade"
        validator = self._get_validator(self.body)
        self.assertTrue(validator.is_valid(self.body))

    @patch.object(UpgradeMessageSender, 'create', Mock(return_value=Mock()))
    def test_controller_with_no_metadata(self):
        """
        Test the mock controller w/out metadata
        """
        tenant_id = '77889991010'
        instance_id = '27e25b73-88a1-4526-b2b9-919a28b8b33f'
        context = Mock()

        req = Mock()
        req.environ = MagicMock()
        req.environ.get = Mock(return_value=context)

        resp = self.controller.create(req, self.body, tenant_id, instance_id)

        instance_version = self.body["upgrade"]["instance_version"]
        location = self.body["upgrade"]["location"]

        metadata = None
        UpgradeMessageSender.create.assert_called_once_with(
            context, instance_id, instance_version, location, metadata)
        self.assertEqual(202, resp.status)

    @patch.object(UpgradeMessageSender, 'create', Mock(return_value=Mock()))
    def test_controller_with_metadata(self):
        """
        Test the mock controller with metadata
        """
        tenant_id = '77889991010'
        instance_id = '27e25b73-88a1-4526-b2b9-919a28b8b33f'
        context = Mock()

        req = Mock()
        req.environ = MagicMock()
        req.environ.get = Mock(return_value=context)

        # append the body w/ metadata
        self.body["upgrade"]["metadata"] = {
            "config_location": "swift://my.conf.location",
            "is_public": True,
            "is_encypted": True}

        resp = self.controller.create(req, self.body, tenant_id, instance_id)

        instance_version = self.body["upgrade"]["instance_version"]
        location = self.body["upgrade"]["location"]
        metadata = self.body["upgrade"]["metadata"]

        UpgradeMessageSender.create.assert_called_once_with(
            context, instance_id, instance_version, location, metadata)
        self.assertEqual(202, resp.status)

    @patch.object(UpgradeMessageSender, 'create', Mock(return_value=Mock()))
    def test_controller_with_empty_metadata(self):
        """
        Test the mock controller with metadata
        """
        tenant_id = '77889991010'
        instance_id = '27e25b73-88a1-4526-b2b9-919a28b8b33f'
        context = Mock()

        req = Mock()
        req.environ = MagicMock()
        req.environ.get = Mock(return_value=context)

        # append the body w/ empty metadata
        self.body["upgrade"]["metadata"] = {}

        resp = self.controller.create(req, self.body, tenant_id, instance_id)

        instance_version = self.body["upgrade"]['instance_version']
        location = self.body["upgrade"]["location"]
        metadata = self.body["upgrade"]["metadata"]

        UpgradeMessageSender.create.assert_called_once_with(
            context, instance_id, instance_version, location, metadata)
        self.assertEqual(202, resp.status)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cinderclient import client as CinderClient
import glanceclient
from keystoneauth1 import loading
from keystoneauth1 import session
from neutronclient.v2_0 import client as NeutronClient
from novaclient.client import Client as NovaClient
from oslo_log import log as logging
import swiftclient

from trove.common import cfg
from trove.common.clients import normalize_url

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
_SESSION = None
ADMIN_NEUTRON_CLIENT = None
ADMIN_NOVA_CLIENT = None
ADMIN_CINDER_CLIENT = None


def get_keystone_session():
    """Get trove service credential auth session."""
    global _SESSION

    if not _SESSION:
        loader = loading.get_plugin_loader('password')
        auth = loader.load_from_options(
            username=CONF.service_credentials.username,
            password=CONF.service_credentials.password,
            project_name=CONF.service_credentials.project_name,
            user_domain_name=CONF.service_credentials.user_domain_name,
            project_domain_name=CONF.service_credentials.project_domain_name,
            auth_url=CONF.service_credentials.auth_url)
        _SESSION = session.Session(auth=auth)

    return _SESSION


def nova_client_trove_admin(context, region_name=None, password=None):
    """
    Returns a nova client object with the trove admin credentials
    :param context: original context from user request
    :type context: trove.common.context.TroveContext
    :return novaclient: novaclient with trove admin credentials
    :rtype: novaclient.client.Client
    """
    global ADMIN_NOVA_CLIENT

    if ADMIN_NOVA_CLIENT:
        LOG.debug('Re-use admin nova client')
        return ADMIN_NOVA_CLIENT

    ks_session = get_keystone_session()
    ADMIN_NOVA_CLIENT = NovaClient(
        CONF.nova_client_version,
        session=ks_session,
        service_type=CONF.nova_compute_service_type,
        region_name=region_name or CONF.service_credentials.region_name,
        insecure=CONF.nova_api_insecure,
        endpoint_type=CONF.nova_compute_endpoint_type)

    if CONF.nova_compute_url and CONF.service_credentials.project_id:
        ADMIN_NOVA_CLIENT.client.endpoint_override = "%s/%s/" % (
            normalize_url(CONF.nova_compute_url),
            CONF.service_credentials.project_id)

    return ADMIN_NOVA_CLIENT


def cinder_client_trove_admin(context, region_name=None):
    """
    Returns a cinder client object with the trove admin credentials
    :param context: original context from user request
    :type context: trove.common.context.TroveContext
    :return cinderclient: cinderclient with trove admin credentials
    """
    global ADMIN_CINDER_CLIENT

    if ADMIN_CINDER_CLIENT:
        LOG.debug('Re-use admin cinder client')
        return ADMIN_CINDER_CLIENT

    version = CONF.cinder_service_type.split('v')[-1] or '3'

    ks_session = get_keystone_session()
    ADMIN_CINDER_CLIENT = CinderClient.Client(
        version,
        session=ks_session,
        service_type=CONF.cinder_service_type,
        region_name=region_name or CONF.service_credentials.region_name,
        insecure=CONF.cinder_api_insecure,
        endpoint_type=CONF.cinder_endpoint_type,
        additional_headers={'OpenStack-API-Version': 'volumev3 latest'})

    if CONF.cinder_url and CONF.service_credentials.project_id:
        ADMIN_CINDER_CLIENT.client.management_url = "%s/%s/" % (
            normalize_url(CONF.cinder_url),
            CONF.service_credentials.project_id)

    return ADMIN_CINDER_CLIENT


def neutron_client_trove_admin(context, region_name=None):
    """
    Returns a neutron client object with the trove admin credentials
    :param context: original context from user request
    :type context: trove.common.context.TroveContext
    :return neutronclient: neutronclient with trove admin credentials
    """
    global ADMIN_NEUTRON_CLIENT

    if ADMIN_NEUTRON_CLIENT:
        LOG.debug('Re-use admin neutron client')
        return ADMIN_NEUTRON_CLIENT

    ks_session = get_keystone_session()
    ADMIN_NEUTRON_CLIENT = NeutronClient.Client(
        session=ks_session,
        service_type=CONF.neutron_service_type,
        region_name=region_name or CONF.service_credentials.region_name,
        insecure=CONF.neutron_api_insecure,
        endpoint_type=CONF.neutron_endpoint_type)

    if CONF.neutron_url:
        ADMIN_NEUTRON_CLIENT.management_url = CONF.neutron_url

    return ADMIN_NEUTRON_CLIENT


def swift_client_trove_admin(context, region_name=None):
    ks_session = get_keystone_session()
    client = swiftclient.Connection(
        session=ks_session,
        insecure=CONF.swift_api_insecure,
        os_options={
            'region_name': region_name or CONF.service_credentials.region_name,
            'service_type': CONF.swift_service_type,
            'endpoint_type': CONF.swift_endpoint_type
        }
    )

    return client


def glance_client_trove_admin(context, region_name=None):
    ks_session = get_keystone_session()
    client = glanceclient.Client(
        version=CONF.glance_client_version,
        session=ks_session,
        region_name=region_name or CONF.service_credentials.region_name,
        service_type=CONF.glance_service_type,
        interface=CONF.glance_endpoint_type
    )

    return client



# Copyright (c) 2018 NEC, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest.mock import Mock
from unittest.mock import patch
from oslo_upgradecheck.upgradecheck import Code

from trove.cmd import status
from trove.tests.unittests import trove_testtools


@patch("trove.cmd.status.db.get_db_api")
@patch("trove.cmd.status.DBInstance")
class TestUpgradeChecksInstancesWithTasks(trove_testtools.TestCase):
    def setUp(self):
        super(TestUpgradeChecksInstancesWithTasks, self).setUp()
        self.cmd = status.Checks()
        self.fake_db_api = Mock()

    def test__check_no_instances_with_tasks(self, mock_instance,
                                            fake_get_db_api):
        fake_get_db_api.return_value = self.fake_db_api

        mock_instance.query.return_value.filter.return_value.filter_by.\
            return_value.count.return_value = 0

        check_result = self.cmd._check_instances_with_running_tasks()
        self.assertEqual(Code.SUCCESS, check_result.code)

    def test__check_instances_with_tasks(self, mock_instance,
                                         fake_get_db_api):
        fake_get_db_api.return_value = self.fake_db_api

        mock_instance.query.return_value.filter.return_value.filter_by.\
            return_value.count.return_value = 1

        check_result = self.cmd._check_instances_with_running_tasks()
        self.assertEqual(Code.WARNING, check_result.code)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy.exc

from trove.common import exception
from trove.db.sqlalchemy import migration
from trove.db.sqlalchemy import session


def list(query_func, *args, **kwargs):
    return query_func(*args, **kwargs).all()


def count(query, *args, **kwargs):
    return query(*args, **kwargs).count()


def first(query, *args, **kwargs):
    return query(*args, **kwargs).first()


def join(query, model, *args):
    return query(model).join(*args)


def find_all(model, **conditions):
    return _query_by(model, **conditions)


def find_all_by_limit(query_func, model, conditions, limit, marker=None,
                      marker_column=None):
    return _limits(query_func, model, conditions, limit, marker,
                   marker_column).all()


def find_by(model, **kwargs):
    return _query_by(model, **kwargs).first()


def find_by_filter(model, **kwargs):
    filters = kwargs.pop('filters', [])
    return _query_by_filter(model, *filters, **kwargs)


def save(model):
    try:
        db_session = session.get_session()
        model = db_session.merge(model)
        db_session.flush()
        return model
    except sqlalchemy.exc.IntegrityError as error:
        raise exception.DBConstraintError(model_name=model.__class__.__name__,
                                          error=str(error.orig))


def delete(model):
    db_session = session.get_session()
    model = db_session.merge(model)
    db_session.delete(model)
    db_session.flush()


def delete_all(query_func, model, **conditions):
    query_func(model, **conditions).delete()


def update(model, **values):
    for k, v in values.items():
        model[k] = v


def update_all(query_func, model, conditions, values):
    query_func(model, **conditions).update(values)


def configure_db(options, *plugins):
    session.configure_db(options)
    configure_db_for_plugins(options, *plugins)


def configure_db_for_plugins(options, *plugins):
    for plugin in plugins:
        session.configure_db(options, models_mapper=plugin.mapper)


def drop_db(options):
    session.drop_db(options)


def clean_db():
    session.clean_db()


def db_sync(options, version=None, repo_path=None):
    migration.db_sync(options, version, repo_path)


def db_upgrade(options, version=None, repo_path=None):
    migration.upgrade(options, version, repo_path)


def db_reset(options, *plugins):
    drop_db(options)
    db_sync(options)
    configure_db(options)


def _base_query(cls):
    return session.get_session().query(cls)


def _query_by(cls, **conditions):
    query = _base_query(cls)
    if conditions:
        query = query.filter_by(**conditions)
    return query


def _query_by_filter(cls, *filters, **conditions):
    query = _query_by(cls, **conditions)
    if filters:
        query = query.filter(*filters)
    return query


def _limits(query_func, model, conditions, limit, marker, marker_column=None):
    query = query_func(model, **conditions)
    marker_column = marker_column or model.id
    if marker:
        query = query.filter(marker_column > marker)
    return query.order_by(marker_column).limit(limit)
